# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A collections of libraries to setup Calibrations.

Note that the set of available libraries will be extended in future releases.
"""

from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, Dict, List, Optional, Set
from warnings import warn
import numpy as np

from qiskit.circuit import Parameter
from qiskit import pulse
from qiskit.pulse import ScheduleBlock
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.calibration_management.calibration_key_types import DefaultCalValue
from qiskit_experiments.exceptions import CalibrationError


class BasisGateLibrary(ABC, Mapping):
    """A base class for libraries of basis gates to make it easier to setup Calibrations."""

    # Location where default parameter values are stored. These may be updated at construction.
    __default_values__ = {}

    # Parameters that do not belong to a schedule, a set of names
    __parameters_without_schedule__ = set()

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, support for pulse "
            "gate calibrations has been deprecated."
        ),
    )
    def __init__(
        self,
        basis_gates: Optional[List[str]] = None,
        default_values: Optional[Dict] = None,
        **extra_kwargs,
    ):
        """Setup the library.

        Args:
            basis_gates: The basis gates to generate.
            default_values: A dictionary to override library default parameter values.
            extra_kwargs: Extra key-word arguments of the subclasses that are saved to be able
                to reconstruct the library using the :meth:`__init__` method.

        Raises:
            CalibrationError: If on of the given basis gates is not supported by the library.
        """
        # Update the default values.
        self._extra_kwargs = extra_kwargs
        self._default_values = self.__default_values__.copy()
        if default_values is not None:
            self._default_values.update(default_values)

        if basis_gates is None:
            basis_gates = list(self.__supported_gates__)

        for gate in basis_gates:
            if gate not in self.__supported_gates__:
                raise CalibrationError(
                    f"Gate {gate} is not supported by {self.__class__.__name__}. "
                    f"Supported gates are: {self.__supported_gates__}."
                )

        self._schedules = self._build_schedules(set(basis_gates))

    @property
    @abstractmethod
    def __supported_gates__(self) -> Dict[str, int]:
        """Return the supported gates of the library.

        The key is the name of the gate and the value is the number of qubits it applies to.
        """
        raise NotImplementedError

    def __getitem__(self, name: str) -> ScheduleBlock:
        """Return the schedule."""
        if name not in self._schedules:
            raise CalibrationError(f"Gate {name} is not contained in {self.__class__.__name__}.")

        return self._schedules[name]

    def __contains__(self, name: str) -> bool:
        """Check if the basis gate is in the library."""
        return name in self._schedules

    def __hash__(self) -> int:
        """Return the hash of the library by computing the hash of the schedule strings."""
        data_to_hash = []
        for name, schedule in sorted(self._schedules.items()):
            data_to_hash.append((name, str(schedule), self.__supported_gates__[name]))

        return hash(tuple(data_to_hash))

    def __len__(self):
        """The length of the library defined as the number of basis gates."""
        return len(self._schedules)

    def __iter__(self):
        """Return an iterator over the basis gate library."""
        return iter(self._schedules)

    def num_qubits(self, name: str) -> int:
        """Return the number of qubits that the schedule with the given name acts on."""
        return self.__supported_gates__[name]

    @property
    def basis_gates(self) -> List[str]:
        """Return the basis gates supported by the library."""
        return list(self._schedules)

    @abstractmethod
    def default_values(self) -> List[DefaultCalValue]:
        """Return the default values for the parameters.

        Returns
            A list of tuples is returned. These tuples are structured so that instances
            of :class:`.Calibrations` can call :meth:`.Calibrations.add_parameter_value`
            on the tuples.
        """

    @abstractmethod
    def _build_schedules(self, basis_gates: Set[str]) -> Dict[str, ScheduleBlock]:
        """Build the schedules stored in the library.

        This method is called as the last step in the :meth:`__init__`. Subclasses must implement
        :meth:`_build_schedules` to build the schedules of the library based on the inputs given
        to the :meth:`__init__` method.

        Args:
            basis_gates: The set of basis gates to build. These will be the supported gates or
                a subset thereof.

        Returns:
            A dictionary where the keys are the names of the schedules/basis gates and the values
            are the corresponding schedules.
        """

    def config(self) -> Dict[str, Any]:
        """Return the settings used to initialize the library."""

        kwargs = {"basis_gates": self.basis_gates, "default_values": self._default_values}
        kwargs.update(self._extra_kwargs)

        return {
            "class": self.__class__.__name__,
            "kwargs": kwargs,
            "hash": hash(self),
        }

    @classmethod
    def from_config(cls, config: Dict) -> "BasisGateLibrary":
        """Deserialize the library given the input dictionary"""
        library = cls(**config["kwargs"])

        if hash(library) != config["hash"]:
            warn(
                "Deserialized basis gate library's hash does not match the hash of the serialized "
                "library. Typically, the hash changes when the internal structure of the template "
                "schedules has been changed."
            )

        return library

    def __json_encode__(self):
        """Convert to format that can be JSON serialized."""
        return self.config()

    @classmethod
    def __json_decode__(cls, value: Dict[str, Any]) -> "BasisGateLibrary":
        """Load from JSON compatible format."""
        return cls.from_config(value)


class FixedFrequencyTransmon(BasisGateLibrary):
    r"""A library of gates for fixed-frequency superconducting qubit architectures.

    Note that for now this library supports single-qubit gates and will be extended
    in the future.

    Provided gates:
        - x: :math:`\pi` pulse around the x-axis.
        - sx: :math:`\pi/2` pulse around the x-axis.
        - y: :math:`\pi` pulse around the y-axis.
        - sy: :math:`\pi/2` pulse around the y-axis.

    Pulse parameters:
        - duration: Duration of the pulses Default value: 160 samples.
        - σ: Standard deviation of the pulses Default value: ``duration / 4``.
        - β: DRAG parameter of the pulses Default value: 0.
        - amp: Magnitude of the complex amplitude of the pulses. If the parameters are
          linked then ``x`` and ``y``
          share the same parameter and ``sx`` and ``sy`` share the same parameter.
          Default value: 50% of the maximum output for ``x`` and ``y`` and 25% of the
          maximum output for ``sx`` and ``sy``. Note that the user provided default amplitude
          in the ``__init__`` method sets the default amplitude of the ``x`` and ``y`` pulses.
          The amplitude of the ``sx`` and ``sy`` pulses is half the provided value.
        - angle: The phase of the complex amplitude of the pulses.

    Parameters without schedule:
        - meas_freq: frequency of the measurement drives.
        - drive_freq: frequency of the qubit drives.

    Note that the β and amp parameters may be linked between the x and y as well as between
    the sx and sy pulses. All pulses share the same duration and σ parameters.
    """

    __default_values__ = {"duration": 160, "amp": 0.5, "β": 0.0, "angle": 0.0}

    __parameters_without_schedule__ = {"meas_freq", "drive_freq"}

    def __init__(
        self,
        basis_gates: Optional[List[str]] = None,
        default_values: Optional[Dict] = None,
        link_parameters: bool = True,
    ):
        """Setup the schedules.

        Args:
            basis_gates: The basis gates to generate.
            default_values: Default values for the parameters this dictionary can contain
                the following keys: "duration", "amp", "β", and "σ". If "σ" is not provided
                this library will take one fourth of the pulse duration as default value.
            link_parameters: If set to ``True``, then the amplitude and DRAG parameters of the
                :math:`X` and :math:`Y` gates will be linked as well as those of
                the :math:`SX` and :math:`SY` gates.
        """
        self._link_parameters = link_parameters

        extra_kwargs = {"link_parameters": link_parameters}

        super().__init__(basis_gates, default_values, **extra_kwargs)

    @property
    def __supported_gates__(self) -> Dict[str, int]:
        """The gates that this library supports."""
        return {"x": 1, "y": 1, "sx": 1, "sy": 1}

    def _build_schedules(self, basis_gates: Set[str]) -> Dict[str, ScheduleBlock]:
        """Build the schedule of the class."""
        dur = Parameter("duration")
        sigma = Parameter("σ")

        x_amp, x_beta, x_angle = Parameter("amp"), Parameter("β"), Parameter("angle")

        if self._link_parameters:
            y_amp, y_beta, y_angle = x_amp, x_beta, x_angle + np.pi / 2
        else:
            y_amp, y_beta, y_angle = Parameter("amp"), Parameter("β"), Parameter("angle")

        sx_amp, sx_beta, sx_angle = Parameter("amp"), Parameter("β"), Parameter("angle")

        if self._link_parameters:
            sy_amp, sy_beta, sy_angle = sx_amp, sx_beta, sx_angle + np.pi / 2
        else:
            sy_amp, sy_beta, sy_angle = Parameter("amp"), Parameter("β"), Parameter("angle")

        # Create the schedules for the gates
        sched_x = self._single_qubit_schedule("x", dur, x_amp, sigma, x_beta, x_angle)
        sched_y = self._single_qubit_schedule("y", dur, y_amp, sigma, y_beta, y_angle)
        sched_sx = self._single_qubit_schedule("sx", dur, sx_amp, sigma, sx_beta, sx_angle)
        sched_sy = self._single_qubit_schedule("sy", dur, sy_amp, sigma, sy_beta, sy_angle)

        schedules = {}
        for sched in [sched_x, sched_y, sched_sx, sched_sy]:
            if sched.name in basis_gates:
                schedules[sched.name] = sched

        return schedules

    @staticmethod
    def _single_qubit_schedule(
        name: str,
        dur: Parameter,
        amp: Parameter,
        sigma: Parameter,
        beta: Parameter,
        angle: Parameter,
    ) -> ScheduleBlock:
        """Build a single qubit pulse."""

        chan = pulse.DriveChannel(Parameter("ch0"))

        with pulse.build(name=name) as sched:
            pulse.play(pulse.Drag(duration=dur, amp=amp, sigma=sigma, beta=beta, angle=angle), chan)

        return sched

    def default_values(self) -> List[DefaultCalValue]:
        """Return the default values for the parameters.

        Returns
            A list of tuples is returned. These tuples are structured so that instances
            of :class:`.Calibrations` can call :meth:`.Calibrations.add_parameter_value`
            on the tuples.
        """
        defaults = []
        for name, schedule in self.items():
            for param in schedule.parameters:
                if "ch" not in param.name:
                    if "y" in name and self._link_parameters:
                        continue

                    if param.name == "σ" and "σ" not in self._default_values:
                        value = self._default_values["duration"] / 4
                    else:
                        value = self._default_values[param.name]

                    if name in {"sx", "sy"} and param.name == "amp":
                        value /= 2.0

                    if "y" in name and param.name == "angle":
                        value += np.pi / 2

                    defaults.append(DefaultCalValue(value, param.name, tuple(), name))

        return defaults


class EchoedCrossResonance(BasisGateLibrary):
    r"""A library for echoed cross-resonance gates.

    The ``cr45p`` and ``cr45m`` include a pulse on the control qubit and optionally a pulse
    on the target qubit.

    Provided gates:
        - cr45p: GaussianSquare cross-resonance gate for a :math:`+\pi/4` rotation.
        - cr45m: GaussianSquare cross-resonance gate for a :math:`-\pi/4` rotation.
        - ecr: Echoed cross-resonance gate defined as ``cr45p - x - cr45m``.
        - rzx: RZXGate built from the ecr as ``cr45p - x - cr45m - x``.

    Required gates:
        - x: the x gate is defined outside of this library, see :class:`.FixedFrequencyTransmon`.

    Pulse parameters:
        - tgt_amp: The amplitude of the pulse applied to the target qubit. Default value: 0.
        - σ: The standard deviation of the flanks. Default value: 64 samples.
        - amp: The amplitude of the pulses applied to the control qubit. Default value: 50%.
        - duration: The duration of the cr45p and cr45m pulses. Default value: 1168 samples.
        - risefall: The number of σ's in the flanks of the pulses. Default value: 2.
    """

    __default_values__ = {
        "tgt_amp": 0.0,
        "tgt_angle": 0.0,
        "amp": 0.5,
        "angle": 0.0,
        "σ": 64,
        "risefall": 2,
        "duration": 1168,
    }

    def __init__(
        self,
        basis_gates: Optional[List[str]] = None,
        default_values: Optional[Dict] = None,
        target_pulses: bool = True,
    ):
        """Setup the library.

        Args:
            basis_gates: The basis gates to generate.
            default_values: A dictionary to override library default parameter values.
            target_pulses: If True (the default) then drives will be added to the target qubit
                during the CR tones on the control qubit.
        """
        self._target_pulses = target_pulses
        super().__init__(basis_gates, default_values)

    @property
    def __supported_gates__(self) -> Dict[str, int]:
        """The supported gates of the library are two-qubit pulses for the ecr gate."""
        return {"cr45p": 2, "cr45m": 2, "ecr": 2, "rzx": 2}

    def default_values(self) -> List[DefaultCalValue]:
        """The default values of the CR library."""
        defaults = []
        for name, schedule in self.items():
            for param in schedule.parameters:
                if "ch" not in param.name:
                    value = self._default_values[param.name]
                    defaults.append(DefaultCalValue(value, param.name, tuple(), name))

        return defaults

    def _build_schedules(self, basis_gates: Set[str]) -> Dict[str, ScheduleBlock]:
        """Build the schedules of the CR library."""

        schedules = {}

        tgt_amp = Parameter("tgt_amp")
        tgt_angle = Parameter("tgt_angle")
        sigma = Parameter("σ")
        cr_amp = Parameter("amp")
        cr_angle = Parameter("angle")
        cr_dur = Parameter("duration")
        cr_rf = Parameter("risefall")
        t_chan_idx = Parameter("ch1")
        u_chan_idx = Parameter("ch0.1")
        t_chan = pulse.DriveChannel(t_chan_idx)
        u_chan = pulse.ControlChannel(u_chan_idx)

        if "cr45p" in basis_gates:
            with pulse.build(name="cr45p") as cr45p:
                pulse.play(
                    pulse.GaussianSquare(
                        cr_dur, cr_amp, angle=cr_angle, risefall_sigma_ratio=cr_rf, sigma=sigma
                    ),
                    u_chan,
                )

                if self._target_pulses:
                    pulse.play(
                        pulse.GaussianSquare(
                            cr_dur,
                            tgt_amp,
                            angle=tgt_angle,
                            risefall_sigma_ratio=cr_rf,
                            sigma=sigma,
                        ),
                        t_chan,
                    )

            schedules["cr45p"] = cr45p

        if "cr45m" in basis_gates:
            with pulse.build(name="cr45m") as cr45m:
                pulse.play(
                    pulse.GaussianSquare(
                        cr_dur,
                        cr_amp,
                        angle=cr_angle + np.pi,
                        risefall_sigma_ratio=cr_rf,
                        sigma=sigma,
                    ),
                    u_chan,
                )

                if self._target_pulses:
                    pulse.play(
                        pulse.GaussianSquare(
                            cr_dur,
                            tgt_amp,
                            angle=tgt_angle + np.pi,
                            risefall_sigma_ratio=cr_rf,
                            sigma=sigma,
                        ),
                        t_chan,
                    )

            schedules["cr45m"] = cr45m

        # Echoed Cross-Resonance gate
        if "ecr" in basis_gates:
            with pulse.build(name="ecr") as ecr:
                with pulse.align_sequential():
                    pulse.reference("cr45p", "q0", "q1")
                    pulse.reference("x", "q0")
                    pulse.reference("cr45m", "q0", "q1")

            schedules["ecr"] = ecr

        # RZXGate built from Echoed Cross-Resonance gate
        if "rzx" in basis_gates:
            with pulse.build(name="rzx") as rzx:
                with pulse.align_sequential():
                    pulse.reference("cr45p", "q0", "q1")
                    pulse.reference("x", "q0")
                    pulse.reference("cr45m", "q0", "q1")
                    pulse.reference("x", "q0")

            schedules["rzx"] = rzx

        return schedules



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Readout Angle Analysis class.
"""

from typing import List, Optional
import numpy as np

from qiskit_experiments.framework import BaseAnalysis, AnalysisResultData, Options
from qiskit_experiments.framework.matplotlib import get_non_gui_ax


class ReadoutAngleAnalysis(BaseAnalysis):
    """
    A class to analyze readout angle experiments
    """

    @classmethod
    def _default_options(cls) -> Options:
        """Return default analysis options.

        Analysis Options:
            plot (bool): Set ``True`` to create figure for fit result.
            ax (AxesSubplot): Optional. A matplotlib axis object to draw.
        """
        options = super()._default_options()
        options.plot = True
        options.ax = None
        return options

    def _run_analysis(self, experiment_data):
        angles = []
        radii = []
        centers = []
        for i in range(2):
            center = complex(*experiment_data.data(i)["memory"][0])
            angles.append(np.angle(center))
            radii.append(np.absolute(center))
            centers.append(center)

        angle = (angles[0] + angles[1]) / 2
        if (np.abs(angles[0] - angles[1])) % (2 * np.pi) > np.pi:
            angle += np.pi

        extra_results = {}
        extra_results["readout_angle_0"] = angles[0]
        extra_results["readout_angle_1"] = angles[1]
        extra_results["readout_radius_0"] = radii[0]
        extra_results["readout_radius_1"] = radii[1]

        analysis_results = [
            AnalysisResultData(name="readout_angle", value=angle, extra=extra_results)
        ]

        if self.options.plot:
            ax = self._format_plot(centers, ax=self.options.ax)
            figures = [ax.get_figure()]
        else:
            figures = None

        return analysis_results, figures

    @staticmethod
    def _format_plot(centers: List[complex], ax: Optional["matplotlib.pyplot.AxesSubplot"] = None):
        """Format the readout_angle plot

        Args:
            centers: the two centers of the level 1 measurements for 0 and for 1.
            ax: matplotlib axis to add plot to.

        Returns:
            AxesSubPlot: the matplotlib axes containing the plot.
        """
        largest_extent = (
            np.max([np.max(np.abs(np.real(centers))), np.max(np.abs(np.imag(centers)))]) * 1.1
        )

        ax = get_non_gui_ax()
        ax.plot(np.real(centers[0]), np.imag(centers[0]), "ro", markersize=24)
        ax.plot(np.real(centers[1]), np.imag(centers[1]), "bo", markersize=24)
        ax.set_xlim([-largest_extent, largest_extent])
        ax.set_ylim([-largest_extent, largest_extent])
        ax.set_xlabel("I [arb.]")
        ax.set_ylabel("Q [arb.]")
        ax.set_title("Centroid Positions")
        ax.legend(["0", "1"])
        return ax



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Curve data classes.
"""
import dataclasses
import itertools
from typing import Any, Dict, Union, List, Tuple, Optional, Iterable

import numpy as np
import uncertainties
from uncertainties.unumpy import uarray

from qiskit_experiments.exceptions import AnalysisError


class CurveFitResult:
    """Result of Qiskit Experiment curve analysis."""

    def __init__(
        self,
        method: Optional[str] = None,
        model_repr: Optional[Dict[str, str]] = None,
        success: Optional[bool] = True,
        nfev: Optional[int] = None,
        message: Optional[str] = "",
        dof: Optional[float] = None,
        init_params: Optional[Dict[str, float]] = None,
        chisq: Optional[float] = None,
        reduced_chisq: Optional[float] = None,
        aic: Optional[float] = None,
        bic: Optional[float] = None,
        params: Optional[Dict[str, float]] = None,
        var_names: Optional[List[str]] = None,
        x_data: Optional[np.ndarray] = None,
        y_data: Optional[np.ndarray] = None,
        weighted_residuals: Optional[np.ndarray] = None,
        residuals: Optional[np.ndarray] = None,
        covar: Optional[np.ndarray] = None,
    ):
        """Create new Qiskit curve analysis result object.

        Args:
            method: A name of fitting algorithm used for the curve fitting.
            model_repr: String representation of fit functions of each curve.
            success: True when the fitting is successfully performed.
            nfev: Number of fit function evaluation until the solution is obtained.
            message: Any message from the fitting software.
            dof: Degree of freedom in this fitting, i.e. number of free parameters.
            init_params: Initial parameters provided to the fitter.
            chisq: Chi-squared value.
            reduced_chisq: Reduced Chi-squared value.
            aic: Akaike's information criterion.
            bic: Bayesian information criterion.
            params: Estimated fitting parameters keyed on the parameter names in the fit function.
            var_names: Name of variables, i.e. fixed parameters are excluded from the list.
            x_data: X values used for the fitting.
            y_data: Y values used for the fitting.
            weighted_residuals: The residuals from the fitting after assigning weights for each ydata.
            residuals: residuals of the fitted model.
            covar: Covariance matrix of fitting variables.
        """
        self.method = method
        self.model_repr = model_repr
        self.success = success
        self.nfev = nfev
        self.message = message
        self.dof = dof
        self.init_params = init_params
        self.chisq = chisq
        self.reduced_chisq = reduced_chisq
        self.aic = aic
        self.bic = bic
        self.params = params
        self.var_names = var_names
        self.x_data = x_data
        self.y_data = y_data
        self.weighted_residuals = weighted_residuals
        self.residuals = residuals
        self.covar = covar

    @property
    def x_range(self) -> Tuple[float, float]:
        """Range of x_data values."""
        return min(self.x_data), max(self.x_data)

    @property
    def y_range(self) -> Tuple[float, float]:
        """Range of y_data values."""
        return min(self.y_data), max(self.y_data)

    @property
    def ufloat_params(self) -> Dict[str, uncertainties.UFloat]:
        """UFloat representation of fit parameters."""
        if hasattr(self, "_ufloat_params"):
            # Return cache
            return getattr(self, "_ufloat_params")

        if self.params is None:
            ufloat_params = None
        else:
            if self.covar is not None:
                ufloat_fitvals = uncertainties.correlated_values(
                    nom_values=[self.params[name] for name in self.var_names],
                    covariance_mat=self.covar,
                    tags=self.var_names,
                )
            else:
                # Invalid covariance matrix. Std dev is set to nan, i.e. not computed.
                with np.errstate(invalid="ignore"):
                    # Setting std_devs to NaN will trigger floating point exceptions
                    # which we can ignore. See https://stackoverflow.com/q/75656026
                    ufloat_fitvals = uarray(
                        nominal_values=[self.params[name] for name in self.var_names],
                        std_devs=np.full(len(self.var_names), np.nan),
                    )
            # Combine fixed params and fitting variables into a single dictionary
            # Fixed parameter has zero std_dev
            ufloat_params = {}
            for name in self.params.keys():
                try:
                    uind = self.var_names.index(name)
                    ufloat_params[name] = ufloat_fitvals[uind]
                except ValueError:
                    ufloat_params[name] = uncertainties.ufloat(self.params[name], std_dev=0.0)

        setattr(self, "_ufloat_params", ufloat_params)
        return ufloat_params

    @property
    def correl(self):
        """Correlation matrix of fit parameters."""
        if hasattr(self, "_correl"):
            # Return cache
            return getattr(self, "_correl")

        if self.covar is not None:
            # This is how uncertainties computes correlation matrix
            stdevs = np.sqrt(np.diag(self.covar))
            correl = self.covar / stdevs / stdevs[:, np.newaxis]
        else:
            correl = None

        setattr(self, "_correl", correl)
        return correl

    def __str__(self):
        ret = "CurveFitResult:"
        ret += f"\n - fitting method: {self.method}"
        ret += f"\n - number of sub-models: {len(self.model_repr)}"
        for model_name, model_expr in self.model_repr.items():
            if len(model_expr) > 60:
                model_expr = f"{model_expr[:60]}..."
            ret += f"\n  * F_{model_name}(x) = {model_expr}"
        ret += f"\n - success: {self.success}"
        ret += f"\n - number of function evals: {self.nfev}"
        ret += f"\n - degree of freedom: {self.dof}"
        ret += f"\n - chi-square: {self.chisq}"
        ret += f"\n - reduced chi-square: {self.reduced_chisq}"
        ret += f"\n - Akaike info crit.: {self.aic}"
        ret += f"\n - Bayesian info crit.: {self.bic}"
        if self.init_params is not None:
            ret += "\n - init params:"
            for name, value in self.init_params.items():
                ret += f"\n  * {name} = {value}"
        if self.ufloat_params is not None:
            ret += "\n - fit params:"
            for name, param in self.ufloat_params.items():
                if np.isfinite(param.std_dev):
                    ret += f"\n  * {name} = {param.nominal_value} ± {param.std_dev}"
                else:
                    ret += f"\n  * {name} = {param.nominal_value}"
        if self.correl is not None:
            ret += "\n - correlations:"
            correlated = {}
            for pi, pj in itertools.combinations(range(len(self.var_names)), 2):
                correlated[(pi, pj)] = self.correl[pi, pj]
            for (pi, pj), corr in sorted(correlated.items(), key=lambda item: item[1]):
                ret += f"\n  * ({self.var_names[pi]}, {self.var_names[pj]}) = {corr}"

        return ret

    def __copy__(self):
        instance = CurveFitResult(**self.__json_encode__())
        # Copying ufloat invalidate parameter correlation.
        # Note that ufloat object has `self._linear_part.linear_combo` dictionary
        # to store parameter correlation keyed on the ufloat objects.
        # Copying the ufloat object may change object id, which is the identifier
        # of ufloat value, thus it invalidates the `linear_combo` dictionary.
        # To avoid missing correlation, the copy invalidate ufloat parameter object cache.
        return instance

    def __deepcopy__(self, memo):
        return self.__copy__()

    def __json_encode__(self):
        return {
            "method": self.method,
            "model_repr": self.model_repr,
            "success": self.success,
            "nfev": self.nfev,
            "message": self.message,
            "dof": self.dof,
            "init_params": self.init_params,
            "chisq": self.chisq,
            "reduced_chisq": self.reduced_chisq,
            "aic": self.aic,
            "bic": self.bic,
            "params": self.params,
            "var_names": self.var_names,
            "x_data": self.x_data,
            "y_data": self.y_data,
            "covar": self.covar,
        }

    @classmethod
    def __json_decode__(cls, value):
        return cls(**value)


@dataclasses.dataclass
class ParameterRepr:
    """Detailed description of fitting parameter.

    Attributes:
        name: Original name of the fit parameter being defined in the fit model.
        repr: Optional. Human-readable parameter name shown in the analysis result and in the figure.
        unit: Optional. Physical unit of this parameter if applicable.
    """

    # Fitter argument name
    name: str

    # Unicode representation
    repr: Optional[str] = None

    # Unit
    unit: Optional[str] = None


class OptionsDict(dict):
    """General extended dictionary for fit options.

    This dictionary provides several extra features.

    - A value setting method which validates the dict key and value.
    - Dictionary keys are limited to those specified in the constructor as ``parameters``.
    """

    def __init__(
        self,
        parameters: List[str],
        defaults: Optional[Union[Iterable[Any], Dict[str, Any]]] = None,
    ):
        """Create new dictionary.

        Args:
            parameters: List of parameter names used in the fit model.
            defaults: Default values.

        Raises:
            AnalysisError: When defaults is provided as array-like but the number of
                element doesn't match with the number of fit parameters.
        """
        if defaults is not None:
            if not isinstance(defaults, dict):
                if len(defaults) != len(parameters):
                    raise AnalysisError(
                        f"Default parameter {defaults} is provided with array-like "
                        "but the number of element doesn't match. "
                        f"This fit requires {len(parameters)} parameters."
                    )
                defaults = dict(zip(parameters, defaults))

            full_options = {p: self.format(defaults.get(p, None)) for p in parameters}
        else:
            full_options = {p: None for p in parameters}

        super().__init__(**full_options)

    def __setitem__(self, key, value):
        """Set value with validations.

        Raises:
            AnalysisError: When key is not previously defined.
        """
        if key not in self:
            raise AnalysisError(f"Parameter {key} is not defined in this fit model.")
        super().__setitem__(key, self.format(value))

    def __hash__(self):
        return hash(tuple(sorted(self.items())))

    def set_if_empty(self, **kwargs):
        """Set value to the dictionary if not assigned.

        Args:
              kwargs: Key and new value to assign.
        """
        for key, value in kwargs.items():
            if self.get(key) is None:
                self[key] = value

    @staticmethod
    def format(value: Any) -> Any:
        """Format dictionary value.

        Subclasses may override this method to provide their own validation.

        Args:
            value: New value to assign.

        Returns:
            Formatted value.
        """
        return value


class InitialGuesses(OptionsDict):
    """Dictionary providing a float validation for initial guesses."""

    @staticmethod
    def format(value: Any) -> Optional[float]:
        """Validate that value is float a float or None.

        Args:
            value: New value to assign.

        Returns:
            Formatted value.

        Raises:
            AnalysisError: When value is not a float or None.
        """
        if value is None:
            return None

        try:
            return float(value)
        except (TypeError, ValueError) as ex:
            raise AnalysisError(f"Input value {value} is not valid initial guess. ") from ex


class Boundaries(OptionsDict):
    """Dictionary providing a validation for boundaries."""

    @staticmethod
    def format(value: Any) -> Optional[Tuple[float, float]]:
        """Validate if value is a min-max value tuple.

        Args:
            value: New value to assign.

        Returns:
            Formatted value.

        Raises:
            AnalysisError: When value is invalid format.
        """
        if value is None:
            return None

        try:
            minv, maxv = value
            if minv >= maxv:
                raise AnalysisError(
                    f"The first value is greater than the second value {minv} >= {maxv}."
                )
            return float(minv), float(maxv)
        except (TypeError, ValueError) as ex:
            raise AnalysisError(f"Input boundary {value} is not a min-max value tuple.") from ex


# pylint: disable=invalid-name
class FitOptions:
    """Collection of fitting options.

    This class is initialized with a list of parameter names used in the fit model
    and corresponding default values provided by users.

    This class is hashable, and generates fitter keyword arguments.
    """

    def __init__(
        self,
        parameters: List[str],
        default_p0: Optional[Union[Iterable[float], Dict[str, float]]] = None,
        default_bounds: Optional[Union[Iterable[Tuple], Dict[str, Tuple]]] = None,
        **extra,
    ):
        # These are private members so that user cannot directly override values
        # without implicitly implemented validation logic. No setter will be provided.
        self.__p0 = InitialGuesses(parameters, default_p0)
        self.__bounds = Boundaries(parameters, default_bounds)
        self.__extra = extra

    def __hash__(self):
        return hash((self.__p0, self.__bounds, tuple(sorted(self.__extra.items()))))

    def __eq__(self, other):
        if isinstance(other, FitOptions):
            checks = [
                self.__p0 == other.__p0,
                self.__bounds == other.__bounds,
                self.__extra == other.__extra,
            ]
            return all(checks)
        return False

    def add_extra_options(self, **kwargs):
        """Add more fitter options."""
        self.__extra.update(kwargs)

    def copy(self):
        """Create copy of this option."""
        return FitOptions(
            parameters=list(self.__p0.keys()),
            default_p0=dict(self.__p0),
            default_bounds=dict(self.__bounds),
            **self.__extra,
        )

    @property
    def p0(self) -> InitialGuesses:
        """Return initial guess dictionary."""
        return self.__p0

    @property
    def bounds(self) -> Boundaries:
        """Return bounds dictionary."""
        return self.__bounds

    @property
    def fitter_opts(self) -> Boundaries:
        """Return fitter options dictionary."""
        return self.__extra

    @property
    def options(self):
        """Generate keyword arguments of the curve fitter."""
        bounds = {k: v if v is not None else (-np.inf, np.inf) for k, v in self.__bounds.items()}
        return {"p0": dict(self.__p0), "bounds": bounds, **self.__extra}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=============================================================
Database Service (:mod:`qiskit_experiments.database_service`)
=============================================================

.. currentmodule:: qiskit_experiments.database_service

This subpackage provides database-specific utility functions and exceptions which
are used with the :class:`.ExperimentData` and :class:`.AnalysisResult` classes.

Device Components
=================

.. autosummary::
   :toctree: ../stubs/

   DeviceComponent
   Qubit
   Resonator
   UnknownComponent
   to_component

Exceptions
==========

.. autosummary::
   :toctree: ../stubs/

   ExperimentDataError
   ExperimentEntryExists
   ExperimentEntryNotFound
"""

from .exceptions import ExperimentDataError, ExperimentEntryExists, ExperimentEntryNotFound
from .device_component import DeviceComponent, Qubit, Resonator, UnknownComponent, to_component



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Ramsey XY frequency calibration experiment."""

from typing import Dict, List, Optional, Sequence

from qiskit.circuit import QuantumCircuit
from qiskit.providers.backend import Backend

from qiskit_experiments.framework import ExperimentData
from qiskit_experiments.library.characterization.ramsey_xy import RamseyXY
from qiskit_experiments.calibration_management.calibrations import Calibrations
from qiskit_experiments.calibration_management.update_library import BaseUpdater
from qiskit_experiments.calibration_management.base_calibration_experiment import (
    BaseCalibrationExperiment,
)


class FrequencyCal(BaseCalibrationExperiment, RamseyXY):
    """A qubit frequency calibration experiment based on the Ramsey XY experiment."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "drive_freq",
        delays: Optional[List] = None,
        osc_freq: float = 2e6,
        auto_update: bool = True,
    ):
        """
        Args:
            physical_qubits: Sequence containing the qubit on which to run the
                frequency calibration.
            calibrations: The calibrations instance with the schedules.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter to update in the calibrations.
                This defaults to `drive_freq`.
            delays: The list of delays that will be scanned in the experiment, in seconds.
            osc_freq: A frequency shift in Hz that will be applied by means of
                a virtual Z rotation to increase the frequency of the measured oscillation.
            auto_update: If set to True, which is the default, then the experiment will
                automatically update the frequency in the calibrations.
        """
        super().__init__(
            calibrations,
            physical_qubits,
            backend=backend,
            delays=delays,
            osc_freq=osc_freq,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

    def _metadata(self) -> Dict[str, any]:
        """Add the oscillation frequency of the experiment to the metadata."""
        metadata = super()._metadata()
        metadata["osc_freq"] = self.experiment_options.osc_freq
        metadata["cal_param_value"] = self._cals.get_parameter_value(
            self._param_name,
            self.physical_qubits,
            group=self.experiment_options.group,
        )

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Adds the calibrations to the transpiled circuits."""
        schedule = self._cals.get_schedule("sx", self.physical_qubits)
        circuit.add_calibration("sx", self.physical_qubits, schedule)

    def update_calibrations(self, experiment_data: ExperimentData):
        """Update the frequency using the reported frequency less the imparted oscillation."""

        result_index = self.experiment_options.result_index
        osc_freq = experiment_data.metadata["osc_freq"]
        group = experiment_data.metadata["cal_group"]
        old_freq = experiment_data.metadata["cal_param_value"]

        fit_freq = BaseUpdater.get_value(experiment_data, "freq", result_index)
        new_freq = old_freq + fit_freq - osc_freq

        BaseUpdater.add_parameter_value(
            self._cals,
            experiment_data,
            new_freq,
            self._param_name,
            group=group,
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
T1 Experiment class.
"""

from typing import List, Optional, Union, Sequence

import numpy as np
from qiskit.circuit import QuantumCircuit
from qiskit.providers.backend import Backend

from qiskit_experiments.framework import BackendTiming, BaseExperiment, Options
from qiskit_experiments.library.characterization.analysis.t1_analysis import T1Analysis


class T1(BaseExperiment):
    r"""An experiment to measure the qubit relaxation time.

    # section: overview

        This experiment estimates the :math:`T_1` relaxation time of the qubit by
        generating a series of circuits that excite the qubit then wait for different
        intervals before measurement. The resulting data of excited population versus
        wait time is fitted to an exponential curve to obtain an estimate for
        :math:`T_1`.

    # section: analysis_ref
        :class:`.T1Analysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_ibm_runtime.fake_provider import FakeManilaV2
            from qiskit_aer import AerSimulator
            backend = AerSimulator.from_backend(FakeManilaV2())

        .. jupyter-execute::

            import numpy as np
            from qiskit_experiments.library import T1

            delays = np.arange(1.e-6, 300.e-6, 30.e-6)
            exp = T1(physical_qubits=(0, ), delays=delays, backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: manual
        :doc:`/manuals/characterization/t1`

    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            delays (Iterable[float]): Delay times of the experiments in seconds.
        """
        options = super()._default_experiment_options()
        options.delays = None
        return options

    def __init__(
        self,
        physical_qubits: Sequence[int],
        delays: Union[List[float], np.array],
        backend: Optional[Backend] = None,
    ):
        """
        Initialize the T1 experiment class.

        Args:
            physical_qubits: a single-element sequence containing the qubit whose T1 is to be
                estimated.
            delays: Delay times of the experiments in seconds.
            backend: Optional, the backend to run the experiment on.

        Raises:
            ValueError: If the number of delays is smaller than 3
        """
        # Initialize base experiment
        super().__init__(physical_qubits, analysis=T1Analysis(), backend=backend)

        # Set experiment options
        self.set_experiment_options(delays=delays)

    def circuits(self) -> List[QuantumCircuit]:
        """
        Return a list of experiment circuits

        Returns:
            The experiment circuits
        """
        timing = BackendTiming(self.backend)

        circuits = []
        for delay in self.experiment_options.delays:
            circ = QuantumCircuit(1, 1)
            circ.x(0)
            circ.barrier(0)
            circ.delay(timing.round_delay(time=delay), 0, timing.delay_unit)
            circ.barrier(0)
            circ.measure(0, 0)

            circ.metadata = {"xval": timing.delay_time(time=delay)}

            circuits.append(circ)

        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Rough amplitude calibration using Rabi."""

from collections import namedtuple
from typing import Dict, Iterable, Optional, Sequence
import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit import Parameter
from qiskit.providers.backend import Backend

from qiskit_experiments.framework import ExperimentData
from qiskit_experiments.calibration_management import BaseCalibrationExperiment, Calibrations
from qiskit_experiments.library.characterization import Rabi
from qiskit_experiments.calibration_management.update_library import BaseUpdater

AnglesSchedules = namedtuple(
    "AnglesSchedules", ["target_angle", "parameter", "schedule", "previous_value"]
)


class RoughAmplitudeCal(BaseCalibrationExperiment, Rabi):
    """A calibration version of the Rabi experiment."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        schedule_name: str = "x",
        amplitudes: Iterable[float] = None,
        cal_parameter_name: Optional[str] = "amp",
        target_angle: float = np.pi,
        auto_update: bool = True,
        group: str = "default",
        backend: Optional[Backend] = None,
    ):
        r"""see class :class:`Rabi` for details.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                rough amplitude calibration.
            calibrations: The calibrations instance with the schedules.
            schedule_name: The name of the schedule to calibrate. Defaults to "x".
            amplitudes: A list of amplitudes to scan. If None is given 51 amplitudes ranging
                from -0.95 to 0.95 will be scanned.
            cal_parameter_name: The name of the parameter in the schedule to update.
            target_angle: The target angle of the gate to calibrate this will default to a
                :math:`\pi`-pulse.
            auto_update: Whether or not to automatically update the calibrations. By
                default this variable is set to True.
            group: The group of calibration parameters to use. The default value is "default".
            backend: Optional, the backend to run the experiment on.
        """
        qubit = physical_qubits[0]
        schedule = calibrations.get_schedule(
            schedule_name, qubit, assign_params={cal_parameter_name: Parameter("amp")}, group=group
        )

        self._validate_channels(schedule, [qubit])
        self._validate_parameters(schedule, 1)

        super().__init__(
            calibrations,
            physical_qubits,
            schedule=schedule,
            amplitudes=amplitudes,
            backend=backend,
            schedule_name=schedule_name,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

        # Set the pulses to update.
        prev_amp = calibrations.get_parameter_value(cal_parameter_name, qubit, schedule_name)
        self.experiment_options.group = group
        self.experiment_options.angles_schedules = [
            AnglesSchedules(
                target_angle=target_angle,
                parameter=cal_parameter_name,
                schedule=schedule_name,
                previous_value=prev_amp,
            )
        ]

    @classmethod
    def _default_experiment_options(cls):
        """Default values for the rough amplitude calibration experiment.

        Experiment Options:
            result_index (int): The index of the result from which to update the calibrations.
            angles_schedules (list(float, str, str, float)): A list of parameter update information.
                Each entry of the list is a tuple with four entries: the target angle of the
                rotation, the name of the amplitude parameter to update, the name of the schedule
                containing the amplitude parameter to update, and the previous value of the
                amplitude parameter to update. This allows one experiment to update several
                schedules, see for example :class:`RoughXSXAmplitudeCal`.
            group (str): The calibration group to which the parameter belongs. This will default
                to the value "default".
        """
        options = super()._default_experiment_options()

        options.angles_schedules = [
            AnglesSchedules(target_angle=np.pi, parameter="amp", schedule="x", previous_value=None)
        ]

        return options

    def _metadata(self) -> Dict[str, any]:
        """Add metadata to the experiment data making it more self contained.

        The following keys are added to each circuit's metadata:
            angles_schedules: A list of parameter update information. Each entry of the list
                is a tuple with four entries: the target angle of the rotation, the name of the
                amplitude parameter to update, the name of the schedule containing the amplitude
                parameter to update, and the previous value of the amplitude parameter to update.
            cal_group: The calibration group to which the amplitude parameters belong.
        """
        metadata = super()._metadata()
        param_values = []
        for angle, param_name, schedule_name, _ in self.experiment_options.angles_schedules:
            param_val = self._cals.get_parameter_value(
                param_name,
                self._physical_qubits,
                schedule_name,
                group=self.experiment_options.group,
            )

            param_values.append(
                AnglesSchedules(
                    target_angle=angle,
                    parameter=param_name,
                    schedule=schedule_name,
                    previous_value=param_val,
                )
            )

        metadata["angles_schedules"] = param_values

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Rabi already has the schedules attached in the program circuits."""
        pass

    def update_calibrations(self, experiment_data: ExperimentData):
        r"""Update the amplitude of one or several schedules.

        The update rule extracts the rate of the oscillation from the fit to the cosine function.
        Recall that the amplitude is the x-axis in the analysis of the :class:`Rabi` experiment.
        The value of the amplitude is thus the desired rotation-angle divided by the rate of
        the oscillation:

        .. math::

            A_i \to \frac{\theta_i}{\omega}

        where :math:`\theta_i` is the desired rotation angle (e.g. :math:`\pi` and :math:`\pi/2`
        for "x" and "sx" gates, respectively) and :math:`\omega` is the rate of the oscillation.

        Args:
            experiment_data: The experiment data from which to extract the measured Rabi oscillation
            used to set the pulse amplitude.
        """

        result_index = self.experiment_options.result_index
        group = experiment_data.metadata["cal_group"]

        rate = 2 * np.pi * BaseUpdater.get_value(experiment_data, self.__outcome__, result_index)

        for angle, param, schedule, prev_amp in experiment_data.metadata["angles_schedules"]:

            # This implementation conserves the type, while working for both real and complex prev_amp
            value = np.round(angle / rate, decimals=8) * prev_amp / np.abs(prev_amp)

            BaseUpdater.add_parameter_value(
                self._cals, experiment_data, value, param, schedule, group
            )


class RoughXSXAmplitudeCal(RoughAmplitudeCal):
    """A rough amplitude calibration of x and sx gates."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        amplitudes: Iterable[float] = None,
        backend: Optional[Backend] = None,
    ):
        """A rough amplitude calibration that updates both the sx and x pulses."""
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name="x",
            amplitudes=amplitudes,
            backend=backend,
            cal_parameter_name="amp",
            target_angle=np.pi,
        )

        self.experiment_options.angles_schedules = [
            AnglesSchedules(target_angle=np.pi, parameter="amp", schedule="x", previous_value=None),
            AnglesSchedules(
                target_angle=np.pi / 2, parameter="amp", schedule="sx", previous_value=None
            ),
        ]


class EFRoughXSXAmplitudeCal(RoughAmplitudeCal):
    r"""A rough amplitude calibration of :math:`X` and :math:`SX` gates on the
    :math:`|1\rangle` <-> :math:`|2\rangle` transition.
    """

    __outcome__ = "rabi_rate_12"

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        amplitudes: Iterable[float] = None,
        backend: Optional[Backend] = None,
        ef_pulse_label: str = "12",
    ):
        r"""A rough amplitude calibration that updates both the sx and x pulses on the
        :math:`|1\rangle` <-> :math:`|2\rangle` transition.

        Args:
            physical_qubits: Sequence containing the index of the qubit
                (technically a qutrit) to run on.
            calibrations: The calibrations instance that stores the pulse schedules.
            amplitudes: The amplitudes to scan.
            backend: Optional, the backend to run the experiment on.
            ef_pulse_label: A label that is post-pended to "x" and "sx" to obtain the name
                of the pulses that drive a :math:`\pi` and :math:`\pi/2` rotation on
                the :math:`|1\rangle` <-> :math:`|2\rangle` transition.
        """
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name="x" + ef_pulse_label,
            amplitudes=amplitudes,
            backend=backend,
            cal_parameter_name="amp",
            target_angle=np.pi,
        )

        self.experiment_options.angles_schedules = [
            AnglesSchedules(
                target_angle=np.pi,
                parameter="amp",
                schedule="x" + ef_pulse_label,
                previous_value=None,
            ),
            AnglesSchedules(
                target_angle=np.pi / 2,
                parameter="amp",
                schedule="sx" + ef_pulse_label,
                previous_value=None,
            ),
        ]

    def _pre_circuit(self) -> QuantumCircuit:
        """A circuit with operations to perform before the Rabi."""
        circ = QuantumCircuit(1)
        circ.x(0)
        return circ

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Attach an x calibration if it is defined."""
        # Attach the x calibration as well if it is in self._cals. We allow for
        # it not to be present in case a user wants to rely on the default x
        # calibration and only calibrate the pulses between levels 1 and 2.
        if self._cals.has_template("x", self.physical_qubits):
            schedule = self._cals.get_schedule("x", self.physical_qubits)
            circuit.add_calibration("x", self.physical_qubits, schedule)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""Actions done on the data to bring it in a usable form.

In Qiskit Experiments, uncertainty propagation is offloaded to the ``uncertainties``
package, that offers a python float and numpy-array compatible number
representation that natively supports standard errors and their propagation.

Given values :math:`a` and :math:`b` with a finite uncertainty, the error propagation
in the function :math:`f` is computed with derivatives

.. math:

    \sigma_f^2 \sim \left| \frac{\partial f}{\partial a} \right|^2 \sigma_a^2
        + \left| \frac{\partial f}{\partial b} \right|^2 \sigma_b^2
        + 2 \frac{\partial f}{\partial a} \frac{\partial f}{\partial b} \sigma_{ab}

where :math:`sigma_a` and :math:`sigma_b` are the uncertainties of :math:`a` and :math:`b` while
:math:`sigma_{ab}` is the correlation between :math:`a` and :math:`b`.
Please refer to the ``uncertainties`` package documentation for additional details.

.. _uncertainties:
https://pypi.org/project/uncertainties/

"""

from typing import Dict, List, Set, Tuple, Union, Any

import numpy as np
from uncertainties import unumpy as unp

from qiskit_experiments.data_processing.data_action import DataAction, TrainableDataAction
from qiskit_experiments.data_processing.exceptions import DataProcessorError


class DataProcessor:
    """
    A DataProcessor defines a sequence of operations to perform on experimental data.
    Calling an instance of DataProcessor applies this sequence on the input argument.
    A DataProcessor is created with a list of DataAction instances. Each DataAction
    applies its ``_process`` method on the data and returns the processed data. The nodes
    in the DataProcessor may also perform data validation and some minor formatting.
    The output of one data action serves as input for the next data action.
    ``DataProcessor.__call__(datum)`` usually takes in an entry from the data property of
    an ExperimentData object (i.e. a dict containing metadata and memory keys and
    possibly counts, like the Result.data property) and produces the formatted data.
    ``DataProcessor.__call__(datum)`` extracts the data from the given datum under
    ``DataProcessor._input_key`` (which is specified at initialization) of the given datum.
    """

    def __init__(
        self,
        input_key: str,
        data_actions: List[DataAction] = None,
    ):
        """Create a chain of data processing actions.

        Args:
            input_key: The initial key in the datum Dict[str, Any] under which the data processor
                will find the data to process.
            data_actions: A list of data processing actions to construct this data processor with.
                If nothing is given the processor returns unprocessed data.
        """
        self._input_key = input_key
        self._nodes = data_actions if data_actions else []

    def append(self, node: DataAction):
        """
        Append new data action node to this data processor.

        Args:
            node: A DataAction that will process the data.
        """
        self._nodes.append(node)

    @property
    def is_trained(self) -> bool:
        """Return True if all nodes of the data processor have been trained."""
        for node in self._nodes:
            if isinstance(node, TrainableDataAction):
                if not node.is_trained:
                    return False

        return True

    def __call__(self, data: Union[Dict, List[Dict]], **options) -> np.ndarray:
        """
        Call self on the given datum. This method sequentially calls the stored data actions
        on the datum.

        Args:
            data: The data, typically from ``ExperimentData.data(...)``,
                that needs to be processed. This dict or list of dicts also contains
                the metadata of each experiment.
            options: Run-time options given as keyword arguments that will be passed to the nodes.

        Returns:
            The data processed by the data processor. This is an arbitrary numpy array that
            may contain standard errors as a ufloat object.
        """
        return self._call_internal(data, **options)

    def call_with_history(
        self, data: Union[Dict, List[Dict]], history_nodes: Set = None
    ) -> Tuple[np.ndarray, List]:
        """
        Call self on the given datum. This method sequentially calls the stored data actions
        on the datum and also returns the history of the processed data.

        Args:
            data: The data, typically from ``ExperimentData.data(...)``,
                that needs to be processed. This dict or list of dicts also contains
                the metadata of each experiment.
            history_nodes: The nodes, specified by index in the data processing chain, to
                include in the history. If None is given then all nodes will be included
                in the history.

        Returns:
            A tuple of (processed data, history), that are the data processed by the processor
            and its intermediate state in each specified node, respectively.
        """
        return self._call_internal(data, True, history_nodes)

    def _call_internal(
        self,
        data: Union[Dict, List[Dict]],
        with_history: bool = False,
        history_nodes: Set = None,
        call_up_to_node: int = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, List]]:
        """Process the data with or without storing the history of the computation.

        Args:
            data: The data, typically from ExperimentData.data(...), that needs to be processed.
            This dict or list of dicts also contains the metadata of each experiment.
            with_history: If True the history is returned otherwise it is not.
            history_nodes: The nodes, specified by index in the data processing chain, to
                include in the history. If None is given then all nodes will be included
                in the history.
            call_up_to_node: The data processor will use each node in the processing chain
                up to the node indexed by call_up_to_node. If this variable is not specified
                then all nodes in the data processing chain will be called.

        Returns:
            When ``with_history`` is ``False`` it returns an numpy array of processed data.
            Otherwise it returns a tuple of (processed data, history) in which the `history`
            is a list of intermediate data at each step.
        """
        if call_up_to_node is None:
            call_up_to_node = len(self._nodes)

        data = self._data_extraction(data)

        history = []
        for index, node in enumerate(self._nodes[:call_up_to_node]):
            data = node(data)

            if with_history and (history_nodes is None or index in history_nodes):
                if data.shape[0] == 1:
                    cache_data = data[0]
                else:
                    cache_data = data
                history.append(
                    (
                        node.__class__.__name__,
                        cache_data,
                        index,
                    )
                )

        # Return only first entry if len(data) == 1, e.g. [[0, 1]] -> [0, 1]
        if data.shape[0] == 1:
            out_data = data[0]
        else:
            out_data = data

        if with_history:
            return out_data, history
        else:
            return out_data

    def train(self, data: Union[Dict, List[Dict]]):
        """Train the nodes of the data processor.

        Args:
            data: The data to use to train the data processor.
        """
        for index, node in enumerate(self._nodes):
            if isinstance(node, TrainableDataAction):
                if not node.is_trained:
                    # Process the data up to the untrained node.
                    node.train(self._call_internal(data, call_up_to_node=index))

    def _data_extraction(self, data: Union[Dict, List[Dict]]) -> np.ndarray:
        """Extracts the data on which to run the nodes.

        If the datum is a list of dicts then the data under self._input_key is extracted
        from each dict and appended to a list which therefore contains all the data.

        Args:
            data: A list of such dicts where the data is contained under the key
                ``self._input_key``.

        Returns:
            The data formatted in such a way that it is ready to be processed by the nodes.

        Raises:
            DataProcessorError:
                - If the input datum is not a list or a dict.
                - If the input key of the data processor is not contained in the data.
                - If the data processor receives multiple data with different measurement
                  configuration, i.e. Jagged array.
        """
        if isinstance(data, dict):
            data = [data]

        data_to_process = []
        dims = None
        for datum in data:
            try:
                outcome = datum[self._input_key]
            except TypeError as error:
                raise DataProcessorError(
                    f"{self.__class__.__name__} only extracts data from "
                    f"lists or dicts, received {type(data)}."
                ) from error
            except KeyError as error:
                raise DataProcessorError(
                    f"The input key {self._input_key} was not found in the input datum."
                ) from error

            if self._input_key != "counts":
                outcome = np.asarray(outcome)
                # Validate data shape
                if dims is None:
                    dims = outcome.shape
                else:
                    # This is because each data node creates full array of all result data.
                    # Jagged array cannot be numerically operated with numpy array.
                    if outcome.shape != dims:
                        raise DataProcessorError(
                            "Input data is likely a mixture of job results with different "
                            "measurement setup. Data processor doesn't support jagged array."
                        )
            data_to_process.append(outcome)

        data_to_process = np.asarray(data_to_process)

        if data_to_process.dtype in (float, int):
            # Likely level1 or below. Return ufloat array with un-computed std_dev.
            # The output data format is a standard ndarray with dtype=object with
            # arbitrary shape [n_circuits, ...] depending on the measurement setup.
            nominal_values = np.asarray(data_to_process, float)
            with np.errstate(invalid="ignore"):
                # Setting std_devs to NaN will trigger floating point exceptions
                # which we can ignore. See https://stackoverflow.com/q/75656026
                uarray = unp.uarray(
                    nominal_values=nominal_values,
                    std_devs=np.full_like(nominal_values, np.nan, dtype=float),
                )
            return uarray
        else:
            # Likely level2 counts or level2 memory data. Cannot be typecasted to ufloat.
            # The output data format is a standard ndarray with dtype=object with
            # shape [n_circuits] or [n_circuits, n_shots].
            # No error value is bound.
            return np.asarray(data_to_process, dtype=object)

    def __repr__(self):
        """String representation of data processors."""
        names = ", ".join(repr(node) for node in self._nodes)

        return f"{self.__class__.__name__}(input_key={self._input_key}, nodes=[{names}])"

    def __json_encode__(self) -> Dict[str, Any]:
        """Return the config dict for this data processor."""
        return {
            "cls": type(self),
            "input_key": self._input_key,
            "nodes": self._nodes,
        }

    @classmethod
    def __json_decode__(cls, config: Dict[str, Any]) -> "DataProcessor":
        """Initialize a data processor from config dict."""
        try:
            return cls(input_key=config["input_key"], data_actions=config["nodes"])
        except KeyError as ex:
            raise KeyError("Imperfect configuration data. Cannot load this processor.") from ex



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Coefficients characterizing Stark shift."""

from __future__ import annotations
from typing import Any

import numpy as np

from qiskit.providers.backend import Backend
from qiskit_ibm_experiment.service import IBMExperimentService
from qiskit_ibm_experiment.exceptions import IBMApiError

from qiskit_experiments.framework.json import ExperimentDecoder
from qiskit_experiments.framework.backend_data import BackendData
from qiskit_experiments.framework.experiment_data import ExperimentData


class StarkCoefficients:
    """A collection of coefficients characterizing Stark shift."""

    def __init__(
        self,
        pos_coef_o1: float,
        pos_coef_o2: float,
        pos_coef_o3: float,
        neg_coef_o1: float,
        neg_coef_o2: float,
        neg_coef_o3: float,
        offset: float,
    ):
        """Create new coefficients object.

        Args:
            pos_coef_o1: The first order shift coefficient on positive amplitude.
            pos_coef_o2: The second order shift coefficient on positive amplitude.
            pos_coef_o3: The third order shift coefficient on positive amplitude.
            neg_coef_o1: The first order shift coefficient on negative amplitude.
            neg_coef_o2: The second order shift coefficient on negative amplitude.
            neg_coef_o3: The third order shift coefficient on negative amplitude.
            offset: Offset frequency.
        """
        self.pos_coef_o1 = pos_coef_o1
        self.pos_coef_o2 = pos_coef_o2
        self.pos_coef_o3 = pos_coef_o3
        self.neg_coef_o1 = neg_coef_o1
        self.neg_coef_o2 = neg_coef_o2
        self.neg_coef_o3 = neg_coef_o3
        self.offset = offset

    def positive_coeffs(self) -> list[float]:
        """Positive coefficients."""
        return [self.pos_coef_o3, self.pos_coef_o2, self.pos_coef_o1]

    def negative_coeffs(self) -> list[float]:
        """Negative coefficients."""
        return [self.neg_coef_o3, self.neg_coef_o2, self.neg_coef_o1]

    def convert_freq_to_amp(
        self,
        freqs: np.ndarray,
    ) -> np.ndarray:
        """A helper function to convert Stark frequency to amplitude.

        Args:
            freqs: Target frequency shifts to compute required Stark amplitude.

        Returns:
            Estimated Stark amplitudes to induce input frequency shifts.

        Raises:
            ValueError: When amplitude value cannot be solved.
        """
        amplitudes = np.zeros_like(freqs)
        for idx, freq in enumerate(freqs):
            shift = freq - self.offset
            if np.isclose(shift, 0.0):
                amplitudes[idx] = 0.0
                continue
            if shift > 0:
                fit = [*self.positive_coeffs(), -shift]
            else:
                fit = [*self.negative_coeffs(), -shift]
            amp_candidates = np.roots(fit)
            # Because the fit function is third order, we get three solutions here.
            criteria = np.all(
                [
                    # Frequency shift and tone have the same sign by definition
                    np.sign(amp_candidates.real) == np.sign(shift),
                    # Tone amplitude is a real value
                    np.isclose(amp_candidates.imag, 0.0),
                    # The absolute value of tone amplitude must be less than 1.0 + 10 mp
                    np.abs(amp_candidates.real) < 1.0 + 10 * np.finfo(float).eps,
                ],
                axis=0,
            )
            valid_amps = amp_candidates[criteria]
            if len(valid_amps) == 0:
                raise ValueError(f"Stark shift at frequency value of {freq} Hz is not available.")
            if len(valid_amps) > 1:
                # We assume a monotonic trend but sometimes a large third-order term causes
                # inflection point and inverts the trend in larger amplitudes.
                # In this case we would have more than one solution, but we can
                # take the smallest amplitude before reaching to the inflection point.
                before_inflection = np.argmin(np.abs(valid_amps.real))
                valid_amp = float(valid_amps[before_inflection].real)
            else:
                valid_amp = float(valid_amps[0].real)
            amplitudes[idx] = min(valid_amp, 1.0)
        return amplitudes

    def convert_amp_to_freq(
        self,
        amps: np.ndarray,
    ) -> np.ndarray:
        """A helper function to convert Stark amplitude to frequency shift.

        Args:
            amps: Amplitude values to convert into frequency shift.

        Returns:
              Calculated frequency shift at given Stark amplitude.
        """
        pos_fit = np.poly1d([*self.positive_coeffs(), self.offset])
        neg_fit = np.poly1d([*self.negative_coeffs(), self.offset])

        return np.where(amps > 0, pos_fit(amps), neg_fit(amps))

    def find_min_max_frequency(
        self,
        min_amp: float,
        max_amp: float,
    ) -> tuple[float, float]:
        """A helper function to estimate maximum frequency shift within given amplitude budget.

        Args:
            min_amp: Minimum Stark amplitude.
            max_amp: Maximum Stark amplitude.

        Returns:
            Minimum and maximum frequency shift available within the amplitude range.
        """
        trim_amps = []
        for amp in [min_amp, max_amp]:
            if amp > 0:
                fit = self.positive_coeffs()
            else:
                fit = self.negative_coeffs()
            # Solve for inflection points by computing the point where derivative becomes zero.
            solutions = np.roots([deriv * coeff for deriv, coeff in zip((3, 2, 1), fit)])
            inflection_points = solutions[
                (solutions.imag == 0) & (np.sign(solutions) == np.sign(amp))
            ]
            if len(inflection_points) > 0:
                # When multiple inflection points are found, use the most outer one.
                # There could be a small inflection point around amp=0,
                # when the first order term is significant.
                amp = min([amp, max(inflection_points, key=abs)], key=abs)
            trim_amps.append(amp)
        return tuple(self.convert_amp_to_freq(np.asarray(trim_amps)))

    def __str__(self):
        # Short representation for dataframe
        return "StarkCoefficients(...)"

    def __eq__(self, other):
        return all(
            [
                self.pos_coef_o1 == other.pos_coef_o1,
                self.pos_coef_o2 == other.pos_coef_o2,
                self.pos_coef_o3 == other.pos_coef_o3,
                self.neg_coef_o1 == other.neg_coef_o1,
                self.neg_coef_o2 == other.neg_coef_o2,
                self.neg_coef_o3 == other.neg_coef_o3,
                self.offset == other.offset,
            ]
        )

    def __json_encode__(self) -> dict[str, Any]:
        return {
            "class": "StarkCoefficients",
            "data": {
                "pos_coef_o1": self.pos_coef_o1,
                "pos_coef_o2": self.pos_coef_o2,
                "pos_coef_o3": self.pos_coef_o3,
                "neg_coef_o1": self.neg_coef_o1,
                "neg_coef_o2": self.neg_coef_o2,
                "neg_coef_o3": self.neg_coef_o3,
                "offset": self.offset,
            },
        }

    @classmethod
    def __json_decode__(cls, value: dict[str, Any]) -> "StarkCoefficients":
        if not value.get("class", None) == "StarkCoefficients":
            raise ValueError("JSON decoded value for StarkCoefficients is not valid class type.")
        return StarkCoefficients(**value.get("data", {}))


def retrieve_coefficients_from_service(
    service: IBMExperimentService,
    backend_name: str,
    qubit: int,
) -> StarkCoefficients:
    """Retrieve StarkCoefficients object from experiment service.

    Args:
        service: IBM Experiment service instance interfacing with result database.
        backend_name: Name of target backend.
        qubit: Index of qubit.

    Returns:
        StarkCoefficients object.

    Raises:
        RuntimeError: When stark_coefficients entry doesn't exist in the service.
    """
    try:
        retrieved = service.analysis_results(
            device_components=[f"Q{qubit}"],
            result_type="stark_coefficients",
            backend_name=backend_name,
            sort_by=["creation_datetime:desc"],
            json_decoder=ExperimentDecoder,
            # Returns the latest value only. IBM service returns 10 entries by default.
            # This could contain old data from previous version, which might not be deserialized.
            limit=1,
        )
    except (IBMApiError, ValueError) as ex:
        raise RuntimeError(
            f"Failed to retrieve the result of stark_coefficients: {ex.message}"
        ) from ex
    if len(retrieved) == 0:
        raise RuntimeError(
            "Analysis result of stark_coefficients is not found in the "
            "experiment service. Run and save the result of StarkRamseyXYAmpScan."
        )

    result_data_dict = retrieved[0].result_data
    if "_value" in result_data_dict:
        # In IBM Experiment service, the result_data["value"] returns
        # a display value for the experiment service webpage.
        # Original data is stored in "_value".
        # TODO: this must be handled by experiment service.
        return result_data_dict["_value"]
    return result_data_dict["value"]


def retrieve_coefficients_from_backend(
    backend: Backend,
    qubit: int,
) -> StarkCoefficients:
    """Retrieve StarkCoefficients object from the Qiskit backend.

    Args:
        backend: Qiskit backend object.
        qubit: Index of qubit.

    Returns:
        StarkCoefficients object.

    Raises:
        RuntimeError: When experiment service cannot be loaded from backend.
    """
    name = BackendData(backend).name
    service = ExperimentData.get_service_from_backend(backend)

    if service is None:
        raise RuntimeError(f"Valid experiment service is not found for the backend {name}.")

    return retrieve_coefficients_from_service(service, name, qubit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Composite Experiment abstract base class.
"""

from typing import List, Sequence, Optional, Union
from abc import abstractmethod
import warnings
from qiskit.providers.backend import Backend
from qiskit_experiments.exceptions import QiskitError
from qiskit_experiments.framework import BaseExperiment
from .composite_analysis import CompositeAnalysis


class CompositeExperiment(BaseExperiment):
    """Composite Experiment base class"""

    def __init__(
        self,
        experiments: List[BaseExperiment],
        physical_qubits: Sequence[int],
        backend: Optional[Backend] = None,
        experiment_type: Optional[str] = None,
        flatten_results: bool = True,
        analysis: Optional[CompositeAnalysis] = None,
    ):
        """Initialize the composite experiment object.

        Args:
            experiments: a list of experiment objects.
            physical_qubits: list of physical qubits for the experiment.
            backend: Optional, the backend to run the experiment on.
            experiment_type: Optional, composite experiment subclass name.
            flatten_results: If True flatten all component experiment results
                             into a single ExperimentData container, including
                             nested composite experiments. If False save each
                             component experiment results as a separate child
                             ExperimentData container. This kwarg is ignored
                             if the analysis kwarg is used.
            analysis: Optional, the composite analysis class to use. If not
                      provided this will be initialized automatically from the
                      supplied experiments.

        Raises:
            QiskitError: If the provided analysis class is not a CompositeAnalysis
                         instance.
        """
        self._experiments = experiments
        self._num_experiments = len(experiments)
        if analysis is None:
            analysis = CompositeAnalysis(
                [exp.analysis for exp in self._experiments], flatten_results=flatten_results
            )
        super().__init__(
            physical_qubits,
            analysis=analysis,
            backend=backend,
            experiment_type=experiment_type,
        )

    @abstractmethod
    def circuits(self):
        pass

    def set_transpile_options(self, **fields):
        super().set_transpile_options(**fields)
        # Recursively set transpile options of component experiments
        for exp in self._experiments:
            exp.set_transpile_options(**fields)

    @property
    def num_experiments(self):
        """Return the number of sub experiments"""
        return self._num_experiments

    def component_experiment(self, index=None) -> Union[BaseExperiment, List[BaseExperiment]]:
        """Return the component Experiment object.

        Args:
            index (int): Experiment index, or ``None`` if all experiments are to be returned.
        Returns:
            BaseExperiment: The component experiment(s).
        """
        if index is None:
            return self._experiments
        return self._experiments[index]

    @property
    def analysis(self) -> Union[CompositeAnalysis, None]:
        """Return the analysis instance for the experiment"""
        return self._analysis

    @analysis.setter
    def analysis(self, analysis: Union[CompositeAnalysis, None]) -> None:
        """Set the analysis instance for the experiment"""
        if analysis is not None and not isinstance(analysis, CompositeAnalysis):
            raise TypeError("Input is not a None or a CompositeAnalysis subclass.")
        self._analysis = analysis

    def copy(self) -> "BaseExperiment":
        """Return a copy of the experiment"""
        ret = super().copy()
        # Recursively call copy of component experiments
        ret._experiments = [exp.copy() for exp in self._experiments]

        # Check if the analysis in CompositeAnalysis was a reference to the
        # original component experiment analyses and if so update the copies
        # to preserve this relationship
        if isinstance(self.analysis, CompositeAnalysis):
            for i, orig_exp in enumerate(self._experiments):
                if orig_exp.analysis is self.analysis._analyses[i]:
                    # Update copies analysis with reference to experiment analysis
                    ret.analysis._analyses[i] = ret._experiments[i].analysis
        return ret

    def set_run_options(self, **fields):
        super().set_run_options(**fields)
        for subexp in self._experiments:
            subexp.set_run_options(**fields)

    def _set_backend(self, backend):
        super()._set_backend(backend)
        for subexp in self._experiments:
            subexp._set_backend(backend)

    def _finalize(self):
        # NOTE: When CompositeAnalysis is updated to support level-1
        # measurements this method should be updated to validate that all
        # sub-experiments have the same meas level and meas return types,
        # and update the composite experiment run option to that value.
        #
        # In addition, we raise an error if we detect inconsistencies in
        # the usage of BatchExperiment separate_job experiment option.

        for i, subexp in enumerate(self._experiments):
            # Validate set and default run options in component experiment
            # against and component experiment run options and raise a
            # warning if any are different and will be overridden
            overridden_keys = []
            sub_vals = []
            comp_vals = []
            for key, sub_val in subexp.run_options.__dict__.items():
                comp_val = getattr(self.run_options, key, None)
                if sub_val != comp_val:
                    overridden_keys.append(key)
                    sub_vals.append(sub_val)
                    comp_vals.append(comp_val)

            if overridden_keys:
                warnings.warn(
                    f"Component {i} {subexp.experiment_type} experiment run options"
                    f" {overridden_keys} values {sub_vals} will be overridden with"
                    f" {self.experiment_type} values {comp_vals}.",
                    UserWarning,
                )
                # Update sub-experiment options with actual run option values so
                # they can be used by that sub experiments _finalize method.
                subexp.set_run_options(**dict(zip(overridden_keys, comp_vals)))

            if not self.experiment_options.get(
                "separate_jobs", False
            ) and subexp.experiment_options.get("separate_jobs", False):
                raise QiskitError(
                    "It is not allowed to request to separate jobs in a child experiment,"
                    " if its parent does not separate jobs as well"
                )

            # Call sub-experiments finalize method
            subexp._finalize()

    def _metadata(self):
        """Add component experiment metadata"""
        metadata = super()._metadata()
        metadata["component_types"] = [
            sub_exp.experiment_type for sub_exp in self.component_experiment()
        ]
        metadata["component_metadata"] = [
            sub_exp._metadata() for sub_exp in self.component_experiment()
        ]
        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""P1 spectroscopy analyses."""

from __future__ import annotations

import numpy as np
from uncertainties import unumpy as unp

from qiskit.utils.deprecation import deprecate_func

import qiskit_experiments.data_processing as dp
import qiskit_experiments.visualization as vis
from qiskit_experiments.data_processing.exceptions import DataProcessorError
from qiskit_experiments.framework import BaseAnalysis, ExperimentData, AnalysisResultData, Options
from .coefficients import (
    StarkCoefficients,
    retrieve_coefficients_from_service,
)


class StarkP1SpectAnalysis(BaseAnalysis):
    """Analysis class for StarkP1Spectroscopy.

    # section: overview

        The P1 landscape is hardly predictable because of the random appearance of
        lossy TLS notches, and hence this analysis doesn't provide any
        generic mathematical model to fit the measurement data.
        A developer may subclass this to conduct own analysis.
        The :meth:`StarkP1SpectAnalysis._run_spect_analysis` is a hook method where
        you can define a custom analysis protocol.

        By default, this analysis just visualizes the measured P1 values against Stark tone amplitudes.
        The tone amplitudes can be converted into the amount of Stark shift
        when the calibrated coefficients are provided in the analysis option,
        or the calibration experiment results are available in the result database.

    # section: see_also
        :class:`qiskit_experiments.library.driven_freq_tuning.StarkRamseyXYAmpScan`

    """

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments and related classses "
            "involving pulse gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(self):
        """Initialize the analysis object."""
        # Pass through to parent. This method is only here to be decorated by
        # deprecate_func
        super().__init__()

    @property
    def plotter(self) -> vis.CurvePlotter:
        """Curve plotter instance."""
        return self.options.plotter

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options.

        Analysis Options:
            plotter (Plotter): Plotter to visualize P1 landscape.
            data_processor (DataProcessor): Data processor to compute P1 value.
            stark_coefficients (Union[Dict, str]): Dictionary of Stark shift coefficients to
                convert tone amplitudes into amount of Stark shift. This dictionary must include
                all keys defined in :attr:`.StarkP1SpectAnalysis.stark_coefficients_names`,
                which are calibrated with :class:`.StarkRamseyXYAmpScan`.
                Alternatively, it searches for these coefficients in the result database
                when "latest" is set. This requires having the experiment service set in
                the experiment data to analyze.
            x_key (str): Key of the circuit metadata to represent x value.
        """
        options = super()._default_options()

        p1spect_plotter = vis.CurvePlotter(vis.MplDrawer())
        p1spect_plotter.set_figure_options(
            xlabel="Stark amplitude",
            ylabel="P(1)",
            xscale="quadratic",
        )

        options.update_options(
            plotter=p1spect_plotter,
            data_processor=dp.DataProcessor("counts", [dp.Probability("1")]),
            stark_coefficients=None,
            x_key="xval",
        )
        options.set_validator("stark_coefficients", StarkCoefficients)

        return options

    # pylint: disable=unused-argument
    def _run_spect_analysis(
        self,
        xdata: np.ndarray,
        ydata: np.ndarray,
        ydata_err: np.ndarray,
    ) -> list[AnalysisResultData]:
        """Run further analysis on the spectroscopy data.

        .. note::
            A subclass can overwrite this method to conduct analysis.

        Args:
            xdata: X values. This is either amplitudes or frequencies.
            ydata: Y values. This is P1 values measured at different Stark tones.
            ydata_err: Sampling error of the Y values.

        Returns:
            A list of analysis results.
        """
        return []

    def _run_analysis(
        self,
        experiment_data: ExperimentData,
    ) -> tuple[list[AnalysisResultData], list["matplotlib.figure.Figure"]]:

        x_key = self.options.x_key

        # Get calibrated Stark tone coefficients
        if self.options.stark_coefficients is None and experiment_data.service is not None:
            # Get value from service
            stark_coeffs = retrieve_coefficients_from_service(
                service=experiment_data.service,
                backend_name=experiment_data.backend_name,
                qubit=experiment_data.metadata["physical_qubits"][0],
            )
        else:
            stark_coeffs = self.options.stark_coefficients

        # Compute P1 value and sampling error
        data = experiment_data.data()
        try:
            xdata = np.asarray([datum["metadata"][x_key] for datum in data], dtype=float)
        except KeyError as ex:
            raise DataProcessorError(
                f"X value key {x_key} is not defined in circuit metadata."
            ) from ex
        ydata_ufloat = self.options.data_processor(data)
        ydata = unp.nominal_values(ydata_ufloat)
        ydata_err = unp.std_devs(ydata_ufloat)

        # Convert x-axis of amplitudes into Stark shift by consuming calibrated parameters.
        if isinstance(stark_coeffs, StarkCoefficients):
            xdata = stark_coeffs.convert_amp_to_freq(amps=xdata)
            self.plotter.set_figure_options(
                xlabel="Stark shift",
                xval_unit="Hz",
                xscale="linear",
            )

        # Draw figures and create analysis results.
        self.plotter.set_series_data(
            series_name="stark_p1",
            x_formatted=xdata,
            y_formatted=ydata,
            y_formatted_err=ydata_err,
            x_interp=xdata,
            y_interp=ydata,
        )
        analysis_results = self._run_spect_analysis(xdata, ydata, ydata_err)

        return analysis_results, [self.plotter.figure()]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Multi state discrimination analysis."""

from typing import List, Tuple, TYPE_CHECKING

import matplotlib
import numpy as np

from qiskit.providers.options import Options
from qiskit_experiments.framework import BaseAnalysis, AnalysisResultData, ExperimentData
from qiskit_experiments.data_processing import SkQDA
from qiskit_experiments.visualization import BasePlotter, IQPlotter, MplDrawer, PlotStyle
from qiskit_experiments.framework.package_deps import HAS_SKLEARN

if TYPE_CHECKING:
    from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis


class MultiStateDiscriminationAnalysis(BaseAnalysis):
    r"""This class fits a multi-state discriminator to the data.

    The class will report the configuration of the discriminator in the analysis result as well as
    the fidelity of the discrimination reported as

    .. math::

        F = 1 - \frac{1}{d}\sum{i\neq j}P(i|j)

    Here, :math:`d` is the number of levels that were discriminated while :math:`P(i|j)` is the
    probability of measuring outcome :math:`i` given that state :math:`j` was prepared.

    .. note::
        This class requires that scikit-learn is installed.
    """

    @classmethod
    @HAS_SKLEARN.require_in_call
    def _default_options(cls) -> Options:
        """Return default analysis options.

        Analysis Options:
            plot (bool): Set ``True`` to create figure for fit result.
            plotter (BasePlotter): A plotter instance to visualize the analysis result.
            ax (AxesSubplot): Optional. A matplotlib axis object in which to draw.
            discriminator (BaseDiscriminator): The sklearn discriminator to classify the data.
                The default is a quadratic discriminant analysis.
        """
        options = super()._default_options()
        options.plotter = IQPlotter(MplDrawer())
        options.plotter.set_options(
            discriminator_max_resolution=64,
            style=PlotStyle(figsize=(6, 4), legend_loc=None),
        )
        options.plot = True
        options.ax = None
        from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis

        options.discriminator = SkQDA(QuadraticDiscriminantAnalysis())
        return options

    @property
    def plotter(self) -> BasePlotter:
        """A short-cut to the IQ plotter instance."""
        return self._options.plotter

    def _run_analysis(
        self,
        experiment_data: ExperimentData,
    ) -> Tuple[List[AnalysisResultData], List["matplotlib.figure.Figure"]]:
        """Train a discriminator based on the experiment data.

        Args:
            experiment_data: the data obtained from the experiment

        Returns:
            The configuration of the trained discriminator and the IQ plot and the fidelity of the
            discrimination.
        """

        # number of states and shots
        n_states = len(experiment_data.data())
        num_shots = len(experiment_data.data()[0]["memory"])

        # Process the data and get labels
        data, fit_state = [], []
        for i in range(n_states):
            state_data = []
            for j in range(num_shots):
                state_data.append(experiment_data.data()[i]["memory"][j][0])
            data.append(np.array(state_data))
            fit_state.append(experiment_data.data()[i]["metadata"]["label"])

        # Train a discriminator on the processed data
        discriminator = self.options.discriminator
        discriminator.fit(
            np.concatenate(data),
            np.asarray([[label] * num_shots for label in fit_state]).flatten().transpose(),
        )

        # Calculate fidelity. First we need to calculate P(i|j):= prob. measuring outcome i given
        # state j was prepared
        predicted_data = [discriminator.predict(state_data) for state_data in data]
        # count per prepared state the number of measured states of each kind and calculate the
        # probability of measuring the wrong state
        prob_wrong = 0
        for i in range(n_states):
            counts = [0] * n_states
            for point in predicted_data[i]:
                counts[point] += 1
            for j in range(n_states):
                if j != i:
                    prob_wrong += counts[j] / num_shots

        # calculate the fidelity
        fidelity = 1 - (1 / n_states) * prob_wrong

        # Crate analysis results from the discriminator configuration
        analysis_results = [
            AnalysisResultData(name="discriminator_config", value=discriminator.config()),
            AnalysisResultData(name="fidelity", value=fidelity),
        ]

        figures = []
        if self.options.plot:
            figures.append(self._levels_plot(discriminator, data, fit_state, fidelity).get_figure())

        return analysis_results, figures

    def _levels_plot(self, discriminator, data, fit_state, fidelity) -> matplotlib.figure.Figure:
        """Helper function for plotting IQ plane for different energy levels.

        Args:
            discriminator: the trained discriminator
            data: the training data
            fit_state: the labels
            fidelity: the fidelity of the classification

        Returns:
            The plotted IQ data.
        """
        # create figure labels
        params_dict = {}
        for state in fit_state:
            params_dict[state] = {"label": f"$|{state}\\rangle$"}
        # Update params_dict to contain any existing series_params values,
        # where they have priority over params_dict.
        params_dict.update(self.plotter.figure_options.series_params)
        self.plotter.set_figure_options(series_params=params_dict)

        # calculate centroids
        centroids = [np.mean(x, axis=0) for x in data]

        for p, c, n in zip(data, centroids, fit_state):
            self.plotter.set_series_data(n, points=p, centroid=c)
        self.plotter.set_supplementary_data(discriminator=discriminator, fidelity=fidelity)

        return self.plotter.figure()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Backend data access helper class

Since `BackendV1` and `BackendV2` do not share the same interface, this
class unifies data access for various data fields.
"""
import warnings
from qiskit.providers.models import PulseBackendConfiguration  # pylint: disable=no-name-in-module
from qiskit.providers import BackendV1, BackendV2
from qiskit.utils.deprecation import deprecate_func


class BackendData:
    """Class for providing joint interface for accessing backend data"""

    def __init__(self, backend):
        """Inits the backend and verifies version"""

        self._backend = backend
        self._v1 = isinstance(backend, BackendV1)
        self._v2 = isinstance(backend, BackendV2)

        if self._v2:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message=".*qiskit.qobj.pulse_qobj.*", category=DeprecationWarning
                )
                self._parse_additional_data()

    def _parse_additional_data(self):
        # data specific parsing not done yet in upstream qiskit
        if hasattr(self._backend, "_conf_dict") and self._backend._conf_dict["open_pulse"]:
            if "u_channel_lo" not in self._backend._conf_dict:
                self._backend._conf_dict["u_channel_lo"] = []  # to avoid qiskit bug
            self._pulse_conf = PulseBackendConfiguration.from_dict(self._backend._conf_dict)

    @property
    def name(self):
        """Returns the backend name"""
        if self._v1:
            return self._backend.name()
        elif self._v2:
            return self._backend.name
        return str(self._backend)

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def control_channel(self, qubits):
        """Returns the backend control channel for the given qubits"""
        try:
            if self._v1:
                return self._backend.configuration().control(qubits)
            elif self._v2:
                try:
                    return self._backend.control_channel(qubits)
                except NotImplementedError:
                    return self._pulse_conf.control(qubits)
        except AttributeError:
            return []
        return []

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def drive_channel(self, qubit):
        """Returns the backend drive channel for the given qubit"""
        try:
            if self._v1:
                return self._backend.configuration().drive(qubit)
            elif self._v2:
                try:
                    return self._backend.drive_channel(qubit)
                except NotImplementedError:
                    return self._pulse_conf.drive(qubit)
        except AttributeError:
            return None
        return None

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def measure_channel(self, qubit):
        """Returns the backend measure channel for the given qubit"""
        try:
            if self._v1:
                return self._backend.configuration().measure(qubit)
            elif self._v2:
                try:
                    return self._backend.measure_channel(qubit)
                except NotImplementedError:
                    return self._pulse_conf.measure(qubit)
        except AttributeError:
            return None
        return None

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def acquire_channel(self, qubit):
        """Returns the backend acquire channel for the given qubit"""
        try:
            if self._v1:
                return self._backend.configuration().acquire(qubit)
            elif self._v2:
                try:
                    return self._backend.acquire_channel(qubit)
                except NotImplementedError:
                    return self._pulse_conf.acquire(qubit)
        except AttributeError:
            return None
        return None

    @property
    def granularity(self):
        """Returns the backend's time constraint granularity"""
        try:
            if self._v1:
                return self._backend.configuration().timing_constraints.get("granularity", 1)
            elif self._v2:
                return self._backend.target.granularity
        except AttributeError:
            return 1
        return 1

    @property
    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        is_property=True,
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def min_length(self):
        """Returns the backend's time constraint minimum duration"""
        try:
            if self._v1:
                return self._backend.configuration().timing_constraints.get("min_length", 0)
            elif self._v2:
                return self._backend.target.min_length
        except AttributeError:
            return 0
        return 0

    @property
    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        is_property=True,
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def pulse_alignment(self):
        """Returns the backend's time constraint pulse alignment"""
        try:
            if self._v1:
                return self._backend.configuration().timing_constraints.get("pulse_alignment", 1)
            elif self._v2:
                return self._backend.target.pulse_alignment
        except AttributeError:
            return 1
        return 1

    @property
    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        is_property=True,
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def acquire_alignment(self):
        """Returns the backend's time constraint acquire alignment"""
        try:
            if self._v1:
                return self._backend.configuration().timing_constraints.get("acquire_alignment", 1)
            elif self._v2:
                return self._backend.target.acquire_alignment
        except AttributeError:
            return 1
        return 1

    @property
    def dt(self):
        """Returns the backend's input time resolution"""
        if self._v1:
            try:
                return self._backend.configuration().dt
            except AttributeError:
                return None
        elif self._v2:
            return self._backend.dt
        return None

    @property
    def max_circuits(self):
        """Returns the backend's max experiments value"""
        if self._v1:
            return getattr(self._backend.configuration(), "max_experiments", None)
        elif self._v2:
            return self._backend.max_circuits
        return None

    @property
    def coupling_map(self):
        """Returns the backend's coupling map"""
        if self._v1:
            return getattr(self._backend.configuration(), "coupling_map", [])
        elif self._v2:
            coupling_map = self._backend.coupling_map
            if coupling_map is None:
                return coupling_map
            return list(coupling_map.get_edges())
        return []

    @property
    def version(self):
        """Returns the backend's version"""
        if self._v1:
            return getattr(self._backend, "version", None)
        elif self._v2:
            return self._backend.version
        return None

    @property
    def provider(self):
        """Returns the backend's provider"""
        try:
            if self._v1:
                return self._backend.provider()
            elif self._v2:
                return self._backend.provider
        except AttributeError:
            return None
        return None

    @property
    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        is_property=True,
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def drive_freqs(self):
        """Returns the backend's qubit drive frequencies"""
        if self._v1:
            return getattr(self._backend.defaults(), "qubit_freq_est", [])
        elif self._v2:
            if self._backend.target.qubit_properties is None:
                return []
            return [property.frequency for property in self._backend.target.qubit_properties]
        return []

    @property
    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        is_property=True,
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, utility functions involving "
            "pulse like this one have been deprecated."
        ),
    )
    def meas_freqs(self):
        """Returns the backend's measurement stimulus frequencies.

        .. note::

            The qiskit base classes do not provide this information as a
            standard backend property, but it is available from some providers
            in the data returned by the ``Backend.defaults()`` method.
        """
        if not hasattr(self._backend, "defaults"):
            return []
        return getattr(self._backend.defaults(), "meas_freq_est", [])

    @property
    def num_qubits(self):
        """Returns the backend's number of qubits"""
        if self._v1:
            return self._backend.configuration().num_qubits
        elif self._v2:
            # meas_freq_est is currently not part of the BackendV2
            return self._backend.num_qubits
        return None

    def qubit_t1(self, qubit: int) -> float:
        """Return the T1 value for a qubit from the backend properties

        Args:
            qubit: the qubit index to return T1 for

        Returns:
            The T1 value
        """
        if self._v1:
            return self._backend.properties().qubit_property(qubit)["T1"][0]
        if self._v2:
            return self._backend.qubit_properties(qubit).t1
        return float("nan")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Clifford synthesis plugins for randomized benchmarking
"""
from __future__ import annotations

from typing import Sequence

from qiskit.circuit import QuantumCircuit, Operation
from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary as sel
from qiskit.exceptions import QiskitError
from qiskit.passmanager.flow_controllers import ConditionalController
from qiskit.synthesis.clifford import synth_clifford_full
from qiskit.transpiler import PassManager, CouplingMap, Layout, Target
from qiskit.transpiler.passes import (
    SabreSwap,
    LayoutTransformation,
    BasisTranslator,
    CheckGateDirection,
    GateDirection,
    Optimize1qGatesDecomposition,
)
from qiskit.transpiler.passes.synthesis.plugin import HighLevelSynthesisPlugin


class RBDefaultCliffordSynthesis(HighLevelSynthesisPlugin):
    """Default Clifford synthesis plugin for randomized benchmarking."""

    def run(
        self,
        high_level_object: Operation,
        coupling_map: CouplingMap | None = None,
        target: Target | None = None,
        qubits: Sequence | None = None,
        **options,
    ) -> QuantumCircuit:
        """Run synthesis for the given Clifford.

        Args:
            high_level_object: The operation to synthesize to a
                :class:`~qiskit.circuit.QuantumCircuit` object.
            coupling_map: The reduced coupling map of the backend. For example,
                if physical qubits [5, 6, 7] to be benchmarked is connected
                as 5 - 7 - 6 linearly, the reduced coupling map is 0 - 2 - 1.
            target: A target representing the target backend, which will be ignored in this plugin.
            qubits: List of physical qubits over which the operation is defined,
                which will be ignored in this plugin.
            options: Additional method-specific optional kwargs,
                which must include ``basis_gates``, basis gates to be used for the synthesis.

        Returns:
            The quantum circuit representation of the Operation
            when successful, and ``None`` otherwise.

        Raises:
            QiskitError: If basis_gates is not supplied.
        """
        # synthesize cliffords
        circ = synth_clifford_full(high_level_object)

        # post processing to comply with basis gates and coupling map
        if coupling_map is None:  # Sabre does not work with coupling_map=None
            return circ

        basis_gates = options.get("basis_gates", None)
        if basis_gates is None:
            raise QiskitError("basis_gates are required to run this synthesis plugin")

        basis_gates = list(basis_gates)

        # Run Sabre routing and undo the layout change
        # assuming Sabre routing does not change the initial layout.
        # And then decompose swap gates, fix 2q-gate direction and optimize 1q gates
        initial_layout = Layout.generate_trivial_layout(*circ.qubits)
        undo_layout_change = LayoutTransformation(
            coupling_map=coupling_map, from_layout="final_layout", to_layout=initial_layout
        )

        def _direction_condition(property_set):
            return not property_set["is_direction_mapped"]

        pm = PassManager(
            [
                SabreSwap(coupling_map),
                undo_layout_change,
                BasisTranslator(sel, basis_gates),
                CheckGateDirection(coupling_map),
                ConditionalController(GateDirection(coupling_map), condition=_direction_condition),
                Optimize1qGatesDecomposition(basis=basis_gates),
            ]
        )
        return pm.run(circ)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Layer Fidelity RB Experiment class.
"""
import functools
import logging
from collections import defaultdict
from typing import Union, Iterable, Optional, List, Sequence, Tuple, Dict

from numpy.random import Generator, default_rng
from numpy.random.bit_generator import BitGenerator, SeedSequence

from qiskit.circuit import QuantumCircuit, CircuitInstruction, Barrier, Gate
from qiskit.circuit.library import get_standard_gate_name_mapping
from qiskit.exceptions import QiskitError
from qiskit.providers import BackendV2Converter
from qiskit.providers.backend import Backend, BackendV1, BackendV2
from qiskit.quantum_info import Clifford
from qiskit.pulse.instruction_schedule_map import CalibrationPublisher

from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.framework.configs import ExperimentConfig
from qiskit_experiments.framework.restless_mixin import RestlessMixin

from .clifford_utils import (
    CliffordUtils,
    DEFAULT_SYNTHESIS_METHOD,
    compose_1q,
    compose_2q,
    inverse_1q,
    inverse_2q,
    num_from_2q_circuit,
    _tensor_1q_nums,
    _clifford_1q_int_to_instruction,
    _clifford_2q_int_to_instruction,
    _decompose_clifford_ops,
)
from .layer_fidelity_analysis import LayerFidelityAnalysis

LOG = logging.getLogger(__name__)


GATE_NAME_MAP = get_standard_gate_name_mapping()
NUM_1Q_CLIFFORD = CliffordUtils.NUM_CLIFFORD_1_QUBIT


class LayerFidelity(BaseExperiment, RestlessMixin):
    r"""A holistic benchmarking experiment to characterize the full quality of the devices at scale.

    # section: overview
        Layer Fidelity (LF) is a method to estimate the fidelity of
        a connecting set of two-qubit gates over :math:`N` qubits by measuring gate errors
        using simultaneous direct randomized benchmarking (RB) in disjoint layers.
        LF can easily be expressed as a layer size independent quantity, error per layered gate (EPLG):
        :math:`EPLG = 1 - LF^{1/N_{2Q}}` where :math:`N_{2Q}` is number of 2-qubit gates in the layers.

        Each of the 2-qubit (or 1-qubit) direct RBs yields the decaying probabilities
        to get back to the ground state for an increasing sequence length (i.e. number of layers),
        fits the exponential curve to estimate the decay rate, and calculates
        the process fidelity of the subsystem from the rate.
        LF is calculated as the product of the 2-qubit (or 1-qubit) process fidelities.
        See Ref. [1] for details.

    # section: analysis_ref
        :class:`LayerFidelityAnalysis`

    # section: reference
        .. ref_arxiv:: 1 2311.05933
    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        two_qubit_layers: Sequence[Sequence[Tuple[int, int]]],
        lengths: Iterable[int],
        backend: Optional[Backend] = None,
        num_samples: int = 6,
        seed: Optional[Union[int, SeedSequence, BitGenerator, Generator]] = None,
        two_qubit_gate: Optional[str] = None,
        one_qubit_basis_gates: Optional[Sequence[str]] = None,
    ):
        """Initialize a layer fidelity experiment.

        Args:
            physical_qubits: List of physical qubits for the experiment.
            two_qubit_layers: List of two-qubit gate layers to run on. Each two-qubit
                    gate layer must be given as a list of directed qubit pairs.
            lengths: A list of layer lengths (the number of depth points).
            backend: The backend to run the experiment on. Note that either ``backend`` or
                    ``two_qubit_gate`` and ``one_qubit_basis_gates`` must be set at instantiation.
            num_samples: Number of samples (i.e. circuits) to generate for each layer length.
            seed: Optional, seed used to initialize ``numpy.random.default_rng``.
                  when generating circuits. The ``default_rng`` will be initialized
                  with this seed value every time :meth:~.LayerFidelity.circuits` is called.
            two_qubit_gate: Optional, 2q-gate name (e.g. "cx", "cz", "ecr")
                            of which the two qubit layers consist.
                            If not specified (but ``backend`` is supplied),
                            one of 2q-gates supported in the backend is automatically set.
            one_qubit_basis_gates: Optional, 1q-gates to use for implementing 1q-Clifford operations.
                            If not specified (but ``backend`` is supplied),
                            all 1q-gates supported in the backend are automatically set.

        Raises:
            QiskitError: If any invalid argument is supplied.
        """
        # Compute full layers
        full_layers = []
        for two_q_layer in two_qubit_layers:
            qubits_in_layer = {q for qpair in two_q_layer for q in qpair}
            if len(qubits_in_layer) != 2 * len(two_q_layer):
                raise QiskitError("two_qubit_layers have a layer with gates on non-disjoint qubits")
            for q in qubits_in_layer:
                if q not in physical_qubits:
                    raise QiskitError(f"Qubit {q} in two_qubit_layers is not in physical_qubits")
            layer = two_q_layer + [(q,) for q in physical_qubits if q not in qubits_in_layer]
            full_layers.append(layer)

        # Initialize base experiment
        super().__init__(
            physical_qubits, analysis=LayerFidelityAnalysis(full_layers), backend=backend
        )
        # assert isinstance(backend, BackendV2)

        # Verify parameters
        if len(set(lengths)) != len(lengths):
            raise QiskitError(f"The lengths list {lengths} should not contain duplicate elements.")
        if num_samples <= 0:
            raise QiskitError(f"The number of samples {num_samples} should be positive.")

        if two_qubit_gate is None:
            if self.backend is None:
                raise QiskitError("two_qubit_gate or backend must be supplied.")
            # Try to set default two_qubit_gate from backend
            for op in self.backend.target.operations:
                if isinstance(op, Gate) and op.num_qubits == 2:
                    two_qubit_gate = op.name
                    LOG.info("%s is set for two_qubit_gate", op.name)
                    break
            if not two_qubit_gate:
                raise QiskitError("two_qubit_gate is not provided and failed to set from backend.")
        else:
            if self.backend is None and two_qubit_gate not in GATE_NAME_MAP:
                raise QiskitError(f"Unknown two_qubit_gate: {two_qubit_gate}.")

        if one_qubit_basis_gates is None:
            if self.backend is None:
                raise QiskitError("one_qubit_basis_gates or backend must be supplied.")
            # Try to set default one_qubit_basis_gates from backend
            one_qubit_basis_gates = []
            for op in self.backend.target.operations:
                if isinstance(op, Gate) and op.num_qubits == 1:
                    one_qubit_basis_gates.append(op.name)
            LOG.info("%s is set for one_qubit_basis_gates", str(one_qubit_basis_gates))
            if not one_qubit_basis_gates:
                raise QiskitError(
                    "one_qubit_basis_gates is not provided and failed to set from backend."
                )
        else:
            if self.backend is None:
                for gate in one_qubit_basis_gates:
                    if gate not in GATE_NAME_MAP:
                        raise QiskitError(f"Unknown gate in one_qubit_basis_gates: {gate}.")

        # Set configurable options
        self.set_experiment_options(
            lengths=sorted(lengths),
            num_samples=num_samples,
            seed=seed,
            two_qubit_layers=two_qubit_layers,
            two_qubit_gate=two_qubit_gate,
            one_qubit_basis_gates=tuple(one_qubit_basis_gates),
        )

        # Verify two_qubit_gate and one_qubit_basis_gates
        self.__validate_basis_gates()

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            two_qubit_layers (List[List[Tuple[int, int]]]): List of two-qubit gate layers to run on.
                Each two-qubit gate layer must be given as a list of directed qubit pairs.
            lengths (List[int]): A list of layer lengths.
            num_samples (int): Number of samples to generate for each layer length.
            seed (None or int or SeedSequence or BitGenerator or Generator): A seed
                used to initialize ``numpy.random.default_rng`` when generating circuits.
                The ``default_rng`` will be initialized with this seed value every time
                :meth:`circuits` is called.
            two_qubit_gate (str): Two-qubit gate name (e.g. "cx", "cz", "ecr")
                of which the two qubit layers consist.
            one_qubit_basis_gates (Tuple[str]): One-qubit gates to use for implementing 1q Cliffords.
            clifford_synthesis_method (str): The name of the Clifford synthesis plugin to use
                for building circuits of RB sequences.
        """
        options = super()._default_experiment_options()
        options.update_options(
            lengths=None,
            num_samples=None,
            seed=None,
            two_qubit_layers=None,
            two_qubit_gate=None,
            one_qubit_basis_gates=None,
            clifford_synthesis_method=DEFAULT_SYNTHESIS_METHOD,
        )
        return options

    def set_experiment_options(self, **fields):
        """Set the experiment options.

        Args:
            fields: The fields to update the options

        Raises:
            AttributeError: If the field passed in is not a supported options
        """
        for field in fields:
            if field in {"two_qubit_layers"}:
                if (
                    hasattr(self._experiment_options, field)
                    and self._experiment_options[field] is not None
                ):
                    raise AttributeError(f"Options field {field} is not allowed to update.")
        super().set_experiment_options(**fields)

    @classmethod
    def _default_transpile_options(cls) -> Options:
        """Default transpiler options for transpiling RB circuits."""
        return Options(optimization_level=1)

    def set_transpile_options(self, **fields):
        """Transpile options is not supported for LayerFidelity experiments.

        Raises:
            QiskitError: If `set_transpile_options` is called.
        """
        raise QiskitError(
            "Custom transpile options is not supported for LayerFidelity experiments."
        )

    def _set_backend(self, backend: Backend):
        """Set the backend V2 for RB experiments since RB experiments only support BackendV2.
        If BackendV1 is provided, it is converted to V2 and stored.
        """
        if isinstance(backend, BackendV1):
            super()._set_backend(BackendV2Converter(backend, add_delay=True))
        else:
            super()._set_backend(backend)
        self.__validate_basis_gates()

    def __validate_basis_gates(self) -> None:
        if not self.backend:
            return
        opts = self.experiment_options
        # validate two_qubit_gate if it is set
        if opts.two_qubit_gate:
            if opts.two_qubit_gate not in self.backend.target.operation_names:
                raise QiskitError(f"two_qubit_gate {opts.two_qubit_gate} is not in backend.target")
            for two_q_layer in opts.two_qubit_layers:
                for qpair in two_q_layer:
                    if not self.backend.target.instruction_supported(opts.two_qubit_gate, qpair):
                        raise QiskitError(f"{opts.two_qubit_gate}{qpair} is not in backend.target")
        # validate one_qubit_basis_gates if it is set
        for gate in opts.one_qubit_basis_gates or []:
            if gate not in self.backend.target.operation_names:
                raise QiskitError(f"{gate} in one_qubit_basis_gates is not in backend.target")
        for gate in opts.one_qubit_basis_gates or []:
            for q in self.physical_qubits:
                if not self.backend.target.instruction_supported(gate, (q,)):
                    raise QiskitError(f"{gate}({q}) is not in backend.target")

    def __residual_qubits(self, two_qubit_layer):
        qubits_in_layer = {q for qpair in two_qubit_layer for q in qpair}
        return [q for q in self.physical_qubits if q not in qubits_in_layer]

    def circuits(self) -> List[QuantumCircuit]:
        r"""Return a list of physical circuits to measure layer fidelity.

        Returns:
            A list of :class:`QuantumCircuit`\s.
        """
        return list(self.circuits_generator())

    def circuits_generator(self) -> Iterable[QuantumCircuit]:
        r"""Return a generator of physical circuits to measure layer fidelity.

        Returns:
            A generator of :class:`QuantumCircuit`\s.
        """
        opts = self.experiment_options
        residal_qubits_by_layer = [self.__residual_qubits(layer) for layer in opts.two_qubit_layers]
        rng = default_rng(seed=opts.seed)
        # define functions and variables for speed
        _to_gate_1q = functools.partial(
            _clifford_1q_int_to_instruction,
            basis_gates=opts.one_qubit_basis_gates,
            synthesis_method=opts.clifford_synthesis_method,
        )
        _to_gate_2q = functools.partial(
            _clifford_2q_int_to_instruction,
            basis_gates=(opts.two_qubit_gate,) + opts.one_qubit_basis_gates,
            coupling_tuple=((0, 1),),
            synthesis_method=opts.clifford_synthesis_method,
        )
        if self.backend:
            gate2q = self.backend.target.operation_from_name(opts.two_qubit_gate)
        else:
            gate2q = GATE_NAME_MAP[opts.two_qubit_gate]
        gate2q_cliff = num_from_2q_circuit(Clifford(gate2q).to_circuit())
        # Circuit generation
        num_qubits = max(self.physical_qubits) + 1
        for i_sample in range(opts.num_samples):
            for i_set, (two_qubit_layer, one_qubits) in enumerate(
                zip(opts.two_qubit_layers, residal_qubits_by_layer)
            ):
                num_2q_gates = len(two_qubit_layer)
                num_1q_gates = len(one_qubits)
                composite_qubits = two_qubit_layer + [(q,) for q in one_qubits]
                composite_clbits = [(2 * c, 2 * c + 1) for c in range(num_2q_gates)]
                composite_clbits.extend(
                    [(c,) for c in range(2 * num_2q_gates, 2 * num_2q_gates + num_1q_gates)]
                )
                for length in opts.lengths:
                    circ = QuantumCircuit(num_qubits, num_qubits)
                    barrier_inst = CircuitInstruction(Barrier(num_qubits), circ.qubits)
                    self.__circuit_body(
                        circ,
                        length,
                        two_qubit_layer,
                        one_qubits,
                        rng,
                        _to_gate_1q,
                        _to_gate_2q,
                        gate2q,
                        gate2q_cliff,
                        barrier_inst,
                    )
                    # add the measurements
                    circ._append(barrier_inst)
                    for qubits, clbits in zip(composite_qubits, composite_clbits):
                        circ.measure(qubits, clbits)
                    # store composite structure in metadata
                    circ.metadata = {
                        "experiment_type": "BatchExperiment",
                        "composite_metadata": [
                            {
                                "experiment_type": "ParallelExperiment",
                                "composite_index": list(range(len(composite_qubits))),
                                "composite_metadata": [
                                    {
                                        "experiment_type": "SubLayerFidelity",
                                        "physical_qubits": qpair,
                                        "sample": i_sample,
                                        "xval": length,
                                    }
                                    for qpair in two_qubit_layer
                                ]
                                + [
                                    {
                                        "experiment_type": "SubLayerFidelity",
                                        "physical_qubits": (q,),
                                        "sample": i_sample,
                                        "xval": length,
                                    }
                                    for q in one_qubits
                                ],
                                "composite_qubits": composite_qubits,
                                "composite_clbits": composite_clbits,
                            }
                        ],
                        "composite_index": [i_set],
                    }
                    yield circ

    @staticmethod
    def __circuit_body(
        circ,
        length,
        two_qubit_layer,
        one_qubits,
        rng,
        _to_gate_1q,
        _to_gate_2q,
        gate2q,
        gate2q_cliff,
        barrier_inst,
    ):
        # initialize cliffords and a ciruit (0: identity clifford)
        cliffs_2q = [0] * len(two_qubit_layer)
        cliffs_1q = [0] * len(one_qubits)
        for _ in range(length):
            # sample random 1q-Clifford layer
            for j, qpair in enumerate(two_qubit_layer):
                # sample product of two 1q-Cliffords as 2q interger Clifford
                samples = rng.integers(NUM_1Q_CLIFFORD, size=2)
                cliffs_2q[j] = compose_2q(cliffs_2q[j], _tensor_1q_nums(*samples))
                # For Clifford 1 (x) Clifford 2, in its circuit representation,
                # Clifford 1 acts on the 2nd qubit and Clifford 2 acts on the 1st qubit.
                # That's why the qpair is reversed here.
                for sample, q in zip(samples, reversed(qpair)):
                    circ._append(_to_gate_1q(sample), (circ.qubits[q],), ())
            for k, q in enumerate(one_qubits):
                sample = rng.integers(NUM_1Q_CLIFFORD)
                cliffs_1q[k] = compose_1q(cliffs_1q[k], sample)
                circ._append(_to_gate_1q(sample), (circ.qubits[q],), ())
            circ._append(barrier_inst)
            # add two qubit gates
            for j, qpair in enumerate(two_qubit_layer):
                circ._append(gate2q, tuple(circ.qubits[q] for q in qpair), ())
                cliffs_2q[j] = compose_2q(cliffs_2q[j], gate2q_cliff)
                # TODO: add dd if necessary
            for k, q in enumerate(one_qubits):
                # TODO: add dd if necessary
                pass
            circ._append(barrier_inst)
        # add the last inverse
        for j, qpair in enumerate(two_qubit_layer):
            inv = inverse_2q(cliffs_2q[j])
            circ._append(_to_gate_2q(inv), tuple(circ.qubits[q] for q in qpair), ())
        for k, q in enumerate(one_qubits):
            inv = inverse_1q(cliffs_1q[k])
            circ._append(_to_gate_1q(inv), (circ.qubits[q],), ())
        return circ

    def _transpiled_circuits(self) -> List[QuantumCircuit]:
        """Return a list of experiment circuits, transpiled."""
        transpiled = [_decompose_clifford_ops(circ) for circ in self.circuits()]
        # Set custom calibrations provided in backend
        if isinstance(self.backend, BackendV2):
            instructions = []  # (op_name, qargs) for each element where qargs mean qubit tuple
            for two_qubit_layer in self.experiment_options.two_qubit_layers:
                for qpair in two_qubit_layer:
                    instructions.append((self.experiment_options.two_qubit_gate, tuple(qpair)))
                for q in self.physical_qubits:
                    for gate_1q in self.experiment_options.one_qubit_basis_gates:
                        instructions.append((gate_1q, (q,)))

            common_calibrations = defaultdict(dict)
            for op_name, qargs in instructions:
                inst_prop = self.backend.target[op_name].get(qargs, None)
                if inst_prop is None:
                    continue
                schedule = inst_prop.calibration
                if schedule is None:
                    continue
                publisher = schedule.metadata.get("publisher", CalibrationPublisher.QISKIT)
                if publisher != CalibrationPublisher.BACKEND_PROVIDER:
                    common_calibrations[op_name][(qargs, tuple())] = schedule

            for circ in transpiled:
                circ.calibrations = common_calibrations

        return transpiled

    def _metadata(self):
        metadata = super()._metadata()
        metadata["two_qubit_layers"] = self.experiment_options.two_qubit_layers
        return metadata

    @classmethod
    def from_config(cls, config: Union[ExperimentConfig, Dict]) -> "LayerFidelity":
        """Initialize an experiment from experiment config"""
        if isinstance(config, dict):
            config = ExperimentConfig(**dict)
        ret = cls(*config.args, **config.kwargs)
        if config.run_options:
            ret.set_run_options(**config.run_options)
        return ret



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Spectroscopy analysis class for resonators."""

from typing import List, Optional, Tuple
import numpy as np

from qiskit.utils.deprecation import deprecate_func

import qiskit_experiments.curve_analysis as curve
from qiskit_experiments.framework import AnalysisResultData, ExperimentData
from qiskit_experiments.framework.matplotlib import get_non_gui_ax
from qiskit_experiments.data_processing.nodes import ProjectorType
from qiskit_experiments.database_service.device_component import Resonator


class ResonatorSpectroscopyAnalysis(curve.ResonanceAnalysis):
    """Class to analysis resonator spectroscopy."""

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments and related classses "
            "involving pulse gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

    @classmethod
    def _default_options(cls):
        """Return default analysis options.

        Analysis Options:
            dimensionality_reduction (ProjectorType): Type of the data processor node
                that will reduce the two-dimensional data to one dimension.
            plot_iq_data (bool): Set True to generate IQ plot.
        """
        options = super()._default_options()
        options.dimensionality_reduction = ProjectorType.ABS
        options.result_parameters = [
            curve.ParameterRepr("freq", "res_freq0", "Hz"),
            curve.ParameterRepr("kappa", "kappa", "Hz"),
        ]
        options.plot_iq_data = True
        return options

    def _get_experiment_components(self, experiment_data: ExperimentData):
        """Return resonators as experiment components."""
        return [Resonator(qubit) for qubit in experiment_data.metadata["physical_qubits"]]

    def _run_analysis(
        self, experiment_data: ExperimentData
    ) -> Tuple[List[AnalysisResultData], List["pyplot.Figure"]]:
        """Wrap the analysis to optionally plot the IQ data."""
        analysis_results, figures = super()._run_analysis(experiment_data)

        if self.options.plot_iq_data:
            axis = get_non_gui_ax()
            figure = axis.get_figure()
            # TODO: Move plotting to a new IQPlotter class.
            figure.set_size_inches(*self.plotter.drawer.style["figsize"])

            iqs = []

            for datum in experiment_data.data():
                if "memory" in datum:
                    mem = np.array(datum["memory"])

                    # Average single-shot data.
                    if len(mem.shape) == 3:
                        for idx in range(mem.shape[1]):
                            iqs.append(np.average(mem[:, idx, :], axis=0))
                    else:
                        iqs.append(mem)

            if len(iqs) > 0:
                iqs = np.vstack(iqs)
                axis.scatter(iqs[:, 0], iqs[:, 1], color="b")
                axis.set_xlabel(
                    "In phase [arb. units]", fontsize=self.plotter.drawer.style["axis_label_size"]
                )
                axis.set_ylabel(
                    "Quadrature [arb. units]", fontsize=self.plotter.drawer.style["axis_label_size"]
                )
                axis.tick_params(labelsize=self.plotter.drawer.style["tick_label_size"])
                axis.grid(True)

                figures.append(figure)

        return analysis_results, figures



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Status of experiment execution."""

from __future__ import annotations

import dataclasses
import enum
from threading import Event
from typing import Optional


class ExperimentStatus(enum.Enum):
    """Class for experiment status enumerated type."""

    EMPTY = "experiment data is empty"
    INITIALIZING = "experiment jobs are being initialized"
    VALIDATING = "experiment jobs are validating"
    QUEUED = "experiment jobs are queued"
    RUNNING = "experiment jobs is actively running"
    CANCELLED = "experiment jobs or analysis has been cancelled"
    POST_PROCESSING = "experiment analysis is actively running"
    DONE = "experiment jobs and analysis have successfully run"
    ERROR = "experiment jobs or analysis incurred an error"

    def __json_encode__(self):
        return self.name

    @classmethod
    def __json_decode__(cls, value):
        return cls.__members__[value]  # pylint: disable=unsubscriptable-object


class AnalysisStatus(enum.Enum):
    """Class for analysis callback status enumerated type."""

    QUEUED = "analysis callback is queued"
    RUNNING = "analysis callback is actively running"
    CANCELLED = "analysis callback has been cancelled"
    DONE = "analysis callback has successfully run"
    ERROR = "analysis callback incurred an error"

    def __json_encode__(self):
        return self.name

    @classmethod
    def __json_decode__(cls, value):
        return cls.__members__[value]  # pylint: disable=unsubscriptable-object


@dataclasses.dataclass
class AnalysisCallback:
    """Dataclass for analysis callback status"""

    name: str = ""
    callback_id: str = ""
    status: AnalysisStatus = AnalysisStatus.QUEUED
    error_msg: Optional[str] = None
    event: Event = dataclasses.field(default_factory=Event)

    def __getstate__(self):
        # We need to remove the Event object from state when pickling
        # since events are not pickleable
        state = self.__dict__
        state["event"] = None
        return state

    def __json_encode__(self):
        return self.__getstate__()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019-2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Calibration helper functions"""

from typing import Optional, Set, Tuple
from functools import lru_cache
import re
import rustworkx as rx

from qiskit.circuit import ParameterExpression, Parameter
from qiskit.pulse import ScheduleBlock

from qiskit_experiments.exceptions import CalibrationError
from qiskit_experiments.calibration_management.calibration_key_types import ScheduleKey


# The channel indices need to be parameterized following this regex.
CHANNEL_PATTERN = r"^ch\d+(\.\d+)*(\$\d+){0,1}$"
CHANNEL_PATTERN_REGEX = re.compile(CHANNEL_PATTERN)


def update_schedule_dependency(schedule: ScheduleBlock, dag: rx.PyDiGraph, key: ScheduleKey):
    """Update a DAG of schedule dependencies.

    Args:
        schedule: A ScheduleBlock that potentially has references to other schedules
            that are already present in the dag.
        dag: A directed acyclic graph that encodes schedule dependencies using references.
        key: The schedule key which also contains the qubits.
    """
    # First, check if the node is already in the DAG.
    try:
        # If it already is in the DAG we remove the existing edges and add the new ones later.
        parent_idx = dag.nodes().index(key)
        for successor in dag.successors(parent_idx):
            dag.remove_edge(parent_idx, dag.nodes().index(successor))

    except ValueError:
        # The schedule is not in the DAG: we add a new node.
        parent_idx = dag.add_node(key)

    for reference in schedule.references:
        ref_key = ScheduleKey(reference[0], key.qubits)
        dag.add_edge(parent_idx, _get_node_index(ref_key, dag), None)


def used_in_references(keys: Set[ScheduleKey], dag: rx.PyDiGraph) -> Set[str]:
    """Find all the schedules in the DAG that reference the given schedules.

    Args:
        keys: A list of schedules keys to which references may exist.
        dag: The dag that represents the dependencies between schedule references.

    Returns:
        A set of schedules that reference the given schedules.
    """
    callers = set()

    for key in keys:
        callers.update(dag.nodes()[idx] for idx in rx.ancestors(dag, _get_node_index(key, dag)))

    return set(key.schedule for key in callers)


def _get_node_index(key: ScheduleKey, dag: rx.PyDiGraph) -> int:
    """A helper method to get the node index in the DAG.

    If the given ScheduleKey is not found then we try and get the default schedule with the
    same name. I.e. a key for which the qubits are an empty tuple.

    Args:
        key: The ScheduleKey for which to find a node in the DAG.
        dag: The DAG of schedule dependencies.

    Returns:
        The index of the node in the dag corresponding to schedule key or its default.
    """
    try:
        return dag.nodes().index(key)
    except ValueError:
        default_key = ScheduleKey(key.schedule, tuple())
        return dag.nodes().index(default_key)


def validate_channels(schedule: ScheduleBlock) -> Set[Parameter]:
    """Validate and get the parameters in the channels of the schedule.

    Channels implicitly defined in references are ignored.

    Args:
        schedule: The schedule for which to get the parameters in the channels.

    Returns:
        The set of parameters explicitly defined in the schedule.

    Raises:
        CalibrationError: If a channel is parameterized by more than one parameter.
        CalibrationError: If the parameterized channel index is not formatted properly.
    """
    param_indices = set()

    # Schedules with references do not explicitly have channels. This needs special handling.
    if schedule.is_referenced():
        for block in schedule.blocks:
            if isinstance(block, ScheduleBlock):
                param_indices.update(validate_channels(block))

        return param_indices

    for ch in schedule.channels:
        if isinstance(ch.index, ParameterExpression):
            if len(ch.index.parameters) != 1:
                raise CalibrationError(f"Channel {ch} can only have one parameter.")

            param_indices.add(ch.index)
            if CHANNEL_PATTERN_REGEX.match(ch.index.name) is None:
                raise CalibrationError(
                    f"Parameterized channel must correspond to {CHANNEL_PATTERN}"
                )

    return param_indices


@lru_cache()
def reference_info(
    reference: Tuple[str, ...],
    qubits: Optional[Tuple[int, ...]] = None,
) -> Tuple[str, Tuple[int, ...]]:
    """Extract reference information from the reference tuple.

    Args:
        reference: The reference of a Reference instruction in a ScheduleBlock.
        qubits: Optional argument to reorder the references.

    Returns:
        A string corresponding to the name of the referenced schedule and the qubits that
        this schedule applies to.

    Raises:
        CalibrationError: If ``reference`` is not a tuple.
        CalibrationError: If ``reference`` is not a tuple of reference name and the qubits that
            that the schedule applies to.
    """
    if not isinstance(reference, tuple):
        raise CalibrationError(f"A schedule reference must be a tuple. Found {reference}.")

    ref_schedule_name, ref_qubits = reference[0], reference[1:]

    if not isinstance(ref_schedule_name, str) and not isinstance(ref_qubits, tuple):
        raise CalibrationError(
            f"A schedule reference is a name and qubits tuple. Found {reference}"
        )

    ref_qubits = tuple(int(qubit[1:]) for qubit in ref_qubits)

    # get the qubit indices for which we are getting the schedules
    if qubits is not None and len(qubits) >= len(ref_qubits):
        ref_qubits = tuple(qubits[idx] for idx in ref_qubits)

    return ref_schedule_name, ref_qubits



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Restless mixin class."""

import logging
from typing import Callable, Sequence, Optional
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit.utils.deprecation import deprecate_func

from qiskit.providers import Backend
from qiskit_experiments.framework import Options
from qiskit_experiments.data_processing.data_processor import DataProcessor
from qiskit_experiments.data_processing.exceptions import DataProcessorError
from qiskit_experiments.data_processing.nodes import ProjectorType
from qiskit_experiments.data_processing import nodes
from qiskit_experiments.data_processing.processor_library import get_kerneled_processor
from qiskit_experiments.framework.base_analysis import BaseAnalysis

LOG = logging.getLogger(__name__)


class RestlessMixin:
    """A mixin to facilitate restless experiments.

    This class defines the following methods:

    - :meth:`~.RestlessMixin.enable_restless`
    - :meth:`~.RestlessMixin._get_restless_processor`
    - :meth:`~.RestlessMixin._t1_check`

    A restless enabled experiment is an experiment that can be run in a restless
    measurement setting. In restless measurements, the qubit is not reset after
    each measurement. Instead, the outcome of the previous quantum non-demolition
    measurement is the initial state for the current circuit. Restless measurements
    therefore require special data processing which is provided by sub-classes of
    the :class:`.RestlessNode`. Restless experiments are a fast alternative for
    several calibration and characterization tasks, for details see
    https://arxiv.org/pdf/2202.06981.pdf.

    This class makes it possible for users to enter a restless run mode without having
    to manually set all the required run options and the data processor. The required options
    are ``rep_delay``, ``init_qubits``, ``memory``, and ``meas_level``. Furthermore,
    subclasses can override the :meth:`_get_restless_processor` method if they require more
    complex restless data processing such as two-qubit calibrations. In addition, this
    class makes it easy to determine if restless measurements are supported for a given
    experiment.

    User Manual
        :doc:`/manuals/measurement/restless_measurements`

    """

    analysis: BaseAnalysis
    _default_run_options: Options()
    set_run_options: Callable
    _backend: Backend
    _physical_qubits: Sequence[int]
    _num_qubits: int

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=("Support for restless experiments has been deprecated."),
    )
    def enable_restless(
        self,
        rep_delay: Optional[float] = None,
        override_processor_by_restless: bool = True,
        suppress_t1_error: bool = False,
    ):
        """Enables a restless experiment by setting the restless run options and the
        restless data processor.

        Args:
            rep_delay: The repetition delay. This is the delay between a measurement
                and the subsequent quantum circuit. Since the backends have
                dynamic repetition rates, the repetition delay can be set to a small
                value which is required for restless experiments. Typical values are
                1 us or less.
            override_processor_by_restless: If False, a data processor that is specified in the
                analysis options of the experiment is not overridden by the restless data
                processor. The default is True.
            suppress_t1_error: If True, the default is False, then no error will be raised when
                ``rep_delay`` is larger than the T1 times of the qubits. Instead, a warning will
                be logged as restless measurements may have a large amount of noise.

        Raises:
            DataProcessorError: If the attribute rep_delay_range is not defined for the backend.
            DataProcessorError: If a data processor has already been set but
                override_processor_by_restless is True.
            DataProcessorError: If the experiment analysis does not have the data_processor
                option.
            DataProcessorError: If the rep_delay is equal to or greater than the
                T1 time of one of the physical qubits in the experiment and the flag
                ``ignore_t1_check`` is False.
        """
        try:
            if not rep_delay:
                # BackendV1 only; BackendV2 does not support this
                rep_delay = self._backend.configuration().rep_delay_range[0]
        except AttributeError as error:
            raise DataProcessorError(
                "The restless experiment can not be enabled because "
                "the attribute rep_delay_range is not defined for this backend "
                "and a minimum rep_delay can not be set."
            ) from error

        # Check the rep_delay compared to the T1 time.
        if not self._t1_check(rep_delay):
            msg = (
                f"The specified repetition delay {rep_delay} is equal to or greater "
                f"than the T1 time of one of the physical qubits"
                f"{self._physical_qubits} in the experiment. Consider choosing "
                f"a smaller repetition delay for the restless experiment."
            )

            if suppress_t1_error:
                LOG.warning(msg)
            else:
                raise DataProcessorError(msg)

        # The excited state promotion readout analysis option is set to
        # False because it is not compatible with a restless experiment.
        meas_level = self._default_run_options().get("meas_level", MeasLevel.CLASSIFIED)
        meas_return = self._default_run_options().get("meas_return", MeasReturnType.SINGLE)
        if not self.analysis.options.get("data_processor", None):
            self.set_run_options(
                rep_delay=rep_delay,
                init_qubits=False,
                memory=True,
                meas_level=meas_level,
                meas_return=meas_return,
                use_measure_esp=False,
            )
            if hasattr(self.analysis.options, "data_processor"):
                self.analysis.set_options(
                    data_processor=self._get_restless_processor(meas_level=meas_level)
                )
            else:
                raise DataProcessorError(
                    "The restless data processor can not be set since the experiment analysis"
                    "does not have the data_processor option."
                )
        else:
            if not override_processor_by_restless:
                self.set_run_options(
                    rep_delay=rep_delay,
                    init_qubits=False,
                    memory=True,
                    meas_level=meas_level,
                    meas_return=meas_return,
                    use_measure_esp=False,
                )
            else:
                raise DataProcessorError(
                    "Cannot enable restless. Data processor has already been set and "
                    "override_processor_by_restless is True."
                )

    def _get_restless_processor(self, meas_level: int = MeasLevel.CLASSIFIED) -> DataProcessor:
        """Returns the restless experiments data processor.

        Notes:
            Sub-classes can override this method if they need more complex data processing.
        """
        outcome = self.analysis.options.get("outcome", "1" * self._num_qubits)
        meas_return = self.analysis.options.get("meas_return", MeasReturnType.SINGLE)
        normalize = self.analysis.options.get("normalization", True)
        dimensionality_reduction = self.analysis.options.get(
            "dimensionality_reduction", ProjectorType.SVD
        )

        if meas_level == MeasLevel.KERNELED:
            return get_kerneled_processor(
                dimensionality_reduction, meas_return, normalize, [nodes.RestlessToIQ()]
            )

        else:
            return DataProcessor(
                "memory",
                [
                    nodes.RestlessToCounts(self._num_qubits),
                    nodes.Probability(outcome),
                ],
            )

    def _t1_check(self, rep_delay: float) -> bool:
        """Check that repetition delay < T1 of the physical qubits in the experiment.

        Args:
            rep_delay: The repetition delay. This is the delay between a measurement
                    and the subsequent quantum circuit.

        Returns:
            True if the repetition delay is smaller than the qubit T1 times.

        Raises:
            DataProcessorError: If the T1 values are not defined for the qubits of
                the used backend.
        """

        try:
            t1_values = [
                self._backend_data.qubit_t1(physical_qubit)
                for physical_qubit in self._physical_qubits
            ]

            if all(rep_delay / t1_value < 1.0 for t1_value in t1_values):
                return True
        except AttributeError as error:
            raise DataProcessorError(
                "The restless experiment can not be enabled since "
                "T1 values are not defined for the qubits of the used backend."
            ) from error

        return False
